"""hug/output_format.py

Defines Hug's built-in output formatting methods

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import base64
import mimetypes
import os
import re
import tempfile
from datetime import date, datetime, timedelta
from decimal import Decimal
from functools import wraps
from io import BytesIO
from operator import itemgetter
from uuid import UUID

import falcon
from falcon import HTTP_NOT_FOUND

from hug import introspect
from hug.format import camelcase, content_type
from hug.json_module import json as json_converter

try:
    import numpy
except ImportError:
    numpy = False

IMAGE_TYPES = (
    "png",
    "jpg",
    "bmp",
    "eps",
    "gif",
    "im",
    "jpeg",
    "msp",
    "pcx",
    "ppm",
    "spider",
    "tiff",
    "webp",
    "xbm",
    "cur",
    "dcx",
    "fli",
    "flc",
    "gbr",
    "gd",
    "ico",
    "icns",
    "imt",
    "iptc",
    "naa",
    "mcidas",
    "mpo",
    "pcd",
    "psd",
    "sgi",
    "tga",
    "wal",
    "xpm",
    "svg",
    "svg+xml",
)

VIDEO_TYPES = (
    ("flv", "video/x-flv"),
    ("mp4", "video/mp4"),
    ("m3u8", "application/x-mpegURL"),
    ("ts", "video/MP2T"),
    ("3gp", "video/3gpp"),
    ("mov", "video/quicktime"),
    ("avi", "video/x-msvideo"),
    ("wmv", "video/x-ms-wmv"),
)
RE_ACCEPT_QUALITY = re.compile("q=(?P<quality>[^;]+)")
json_converters = {}
stream = tempfile.NamedTemporaryFile if "UWSGI_ORIGINAL_PROC_NAME" in os.environ else BytesIO


def _json_converter(item):
    if hasattr(item, "__native_types__"):
        return item.__native_types__()

    for kind, transformer in json_converters.items():
        if isinstance(item, kind):
            return transformer(item)

    if isinstance(item, (date, datetime)):
        return item.isoformat()
    elif isinstance(item, bytes):
        try:
            return item.decode("utf8")
        except UnicodeDecodeError:
            return base64.b64encode(item)
    elif hasattr(item, "__iter__"):
        return list(item)
    elif isinstance(item, (Decimal, UUID)):
        return str(item)
    elif isinstance(item, timedelta):
        return item.total_seconds()
    raise TypeError("Type not serializable")


def json_convert(*kinds):
    """Registers the wrapped method as a JSON converter for the provided types.

    NOTE: custom converters are always globally applied
    """

    def register_json_converter(function):
        for kind in kinds:
            json_converters[kind] = function
        return function

    return register_json_converter


if numpy:

    @json_convert(numpy.ndarray)
    def numpy_listable(item):
        return item.tolist()

    @json_convert(str, numpy.unicode_)
    def numpy_stringable(item):
        return str(item)

    @json_convert(numpy.bytes_)
    def numpy_byte_decodeable(item):
        return item.decode()

    @json_convert(numpy.bool_)
    def numpy_boolable(item):
        return bool(item)

    @json_convert(numpy.integer)
    def numpy_integerable(item):
        return int(item)

    @json_convert(float, numpy.floating)
    def numpy_floatable(item):
        return float(item)


@content_type("application/json; charset=utf-8")
def json(content, request=None, response=None, ensure_ascii=False, **kwargs):
    """JSON (Javascript Serialized Object Notation)"""
    if hasattr(content, "read"):
        return content

    if isinstance(content, tuple) and getattr(content, "_fields", None):
        content = {field: getattr(content, field) for field in content._fields}
    return json_converter.dumps(
        content, default=_json_converter, ensure_ascii=ensure_ascii, **kwargs
    ).encode("utf8")


def on_valid(valid_content_type, on_invalid=json):
    """Renders as the specified content type only if no errors are found in the provided data object"""
    invalid_kwargs = introspect.generate_accepted_kwargs(on_invalid, "request", "response")
    invalid_takes_response = introspect.takes_all_arguments(on_invalid, "response")

    def wrapper(function):
        valid_kwargs = introspect.generate_accepted_kwargs(function, "request", "response")
        valid_takes_response = introspect.takes_all_arguments(function, "response")

        @content_type(valid_content_type)
        @wraps(function)
        def output_content(content, response, **kwargs):
            if type(content) == dict and "errors" in content:
                response.content_type = on_invalid.content_type
                if invalid_takes_response:
                    kwargs["response"] = response
                return on_invalid(content, **invalid_kwargs(kwargs))

            if valid_takes_response:
                kwargs["response"] = response
            return function(content, **valid_kwargs(kwargs))

        return output_content

    return wrapper


@content_type("text/plain; charset=utf-8")
def text(content, **kwargs):
    """Free form UTF-8 text"""
    if hasattr(content, "read"):
        return content

    return str(content).encode("utf8")


@content_type("text/html; charset=utf-8")
def html(content, **kwargs):
    """HTML (Hypertext Markup Language)"""
    if hasattr(content, "read"):
        return content
    elif hasattr(content, "render"):
        return content.render().encode("utf8")

    return str(content).encode("utf8")


def _camelcase(content):
    if isinstance(content, dict):
        new_dictionary = {}
        for key, value in content.items():
            if isinstance(key, str):
                key = camelcase(key)
            new_dictionary[key] = _camelcase(value)
        return new_dictionary
    elif isinstance(content, list):
        new_list = []
        for element in content:
            new_list.append(_camelcase(element))
        return new_list
    else:
        return content


@content_type("application/json; charset=utf-8")
def json_camelcase(content, **kwargs):
    """JSON (Javascript Serialized Object Notation) with all keys camelCased"""
    return json(_camelcase(content), **kwargs)


@content_type("application/json; charset=utf-8")
def pretty_json(content, **kwargs):
    """JSON (Javascript Serialized Object Notion) pretty printed and indented"""
    return json(content, indent=4, separators=(",", ": "), **kwargs)


def image(image_format, doc=None):
    """Dynamically creates an image type handler for the specified image type"""

    @on_valid("image/{0}".format(image_format))
    def image_handler(data, **kwargs):
        if hasattr(data, "read"):
            return data
        elif hasattr(data, "save"):
            output = stream()
            if introspect.takes_all_arguments(data.save, "format") or introspect.takes_kwargs(
                data.save
            ):
                data.save(output, format=image_format.upper())
            else:
                data.save(output)
            output.seek(0)
            return output
        elif hasattr(data, "render"):
            return data.render()
        elif os.path.isfile(data):
            return open(data, "rb")

    image_handler.__doc__ = doc or "{0} formatted image".format(image_format)
    return image_handler


for image_type in IMAGE_TYPES:
    globals()["{0}_image".format(image_type.replace("+", "_"))] = image(image_type)


def video(video_type, video_mime, doc=None):
    """Dynamically creates a video type handler for the specified video type"""

    @on_valid(video_mime)
    def video_handler(data, **kwargs):
        if hasattr(data, "read"):
            return data
        elif hasattr(data, "save"):
            output = stream()
            data.save(output, format=video_type.upper())
            output.seek(0)
            return output
        elif hasattr(data, "render"):
            return data.render()
        elif os.path.isfile(data):
            return open(data, "rb")

    video_handler.__doc__ = doc or "{0} formatted video".format(video_type)
    return video_handler


for (video_type, video_mime) in VIDEO_TYPES:
    globals()["{0}_video".format(video_type)] = video(video_type, video_mime)


@on_valid("file/dynamic")
def file(data, response, **kwargs):
    """A dynamically retrieved file"""
    if not data:
        response.content_type = "text/plain"
        return ""

    if hasattr(data, "read"):
        name, data = getattr(data, "name", ""), data
    elif os.path.isfile(data):
        name, data = data, open(data, "rb")
    else:
        response.content_type = "text/plain"
        response.status = HTTP_NOT_FOUND
        return "File not found!"

    response.content_type = mimetypes.guess_type(name, None)[0] or "application/octet-stream"
    return data


def on_content_type(
    handlers, default=None, error="The requested content type does not match any of those allowed"
):
    """Returns a content in a different format based on the clients provided content type,
       should pass in a dict with the following format:

            {'[content-type]': action,
             ...
            }
    """

    def output_type(data, request, response):
        handler = handlers.get(request.content_type.split(";")[0], default)
        if not handler:
            raise falcon.HTTPNotAcceptable(error)

        response.content_type = handler.content_type
        return handler(data, request=request, response=response)

    output_type.__doc__ = "Supports any of the following formats: {0}".format(
        ", ".join(function.__doc__ or function.__name__ for function in handlers.values())
    )
    output_type.content_type = ", ".join(handlers.keys())
    return output_type


def accept_quality(accept, default=1):
    """Separates out the quality score from the accepted content_type"""
    quality = default
    if accept and ";" in accept:
        accept, rest = accept.split(";", 1)
        accept_quality = RE_ACCEPT_QUALITY.search(rest)
        if accept_quality:
            quality = float(accept_quality.groupdict().get("quality", quality).strip())

    return (quality, accept.strip())


def accept(
    handlers, default=None, error="The requested content type does not match any of those allowed"
):
    """Returns a content in a different format based on the clients defined accepted content type,
       should pass in a dict with the following format:

            {'[content-type]': action,
             ...
            }
    """

    def output_type(data, request, response):
        accept = request.accept
        if accept in ("", "*", "/"):
            handler = default or handlers and next(iter(handlers.values()))
        else:
            handler = default
            accepted = [accept_quality(accept_type) for accept_type in accept.split(",")]
            accepted.sort(key=itemgetter(0))
            for _quality, accepted_content_type in reversed(accepted):
                if accepted_content_type in handlers:
                    handler = handlers[accepted_content_type]
                    break

        if not handler:
            raise falcon.HTTPNotAcceptable(error)

        response.content_type = handler.content_type
        return handler(data, request=request, response=response)

    output_type.__doc__ = "Supports any of the following formats: {0}".format(
        ", ".join(function.__doc__ for function in handlers.values())
    )
    output_type.content_type = ", ".join(handlers.keys())
    return output_type


def suffix(
    handlers, default=None, error="The requested suffix does not match any of those allowed"
):
    """Returns a content in a different format based on the suffix placed at the end of the URL route
       should pass in a dict with the following format:

            {'[suffix]': action,
             ...
            }
    """

    def output_type(data, request, response):
        path = request.path
        handler = default
        for suffix_test, suffix_handler in handlers.items():
            if path.endswith(suffix_test):
                handler = suffix_handler
                break

        if not handler:
            raise falcon.HTTPNotAcceptable(error)

        response.content_type = handler.content_type
        return handler(data, request=request, response=response)

    output_type.__doc__ = "Supports any of the following formats: {0}".format(
        ", ".join(function.__doc__ for function in handlers.values())
    )
    output_type.content_type = ", ".join(handlers.keys())
    return output_type


def prefix(
    handlers, default=None, error="The requested prefix does not match any of those allowed"
):
    """Returns a content in a different format based on the prefix placed at the end of the URL route
       should pass in a dict with the following format:

            {'[prefix]': action,
             ...
            }
    """

    def output_type(data, request, response):
        path = request.path
        handler = default
        for prefix_test, prefix_handler in handlers.items():
            if path.startswith(prefix_test):
                handler = prefix_handler
                break

        if not handler:
            raise falcon.HTTPNotAcceptable(error)

        response.content_type = handler.content_type
        return handler(data, request=request, response=response)

    output_type.__doc__ = "Supports any of the following formats: {0}".format(
        ", ".join(function.__doc__ for function in handlers.values())
    )
    output_type.content_type = ", ".join(handlers.keys())
    return output_type



"""hug/use.py

Provides a mechanism for using external hug APIs both locally or remotely in a seamless fashion

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import re
import socket
from collections import namedtuple
from io import BytesIO
from queue import Queue

import falcon
import requests

import hug._empty as empty
from hug.api import API
from hug.defaults import input_format
from hug.format import parse_content_type

Response = namedtuple("Response", ("data", "status_code", "headers"))
Request = namedtuple("Request", ("content_length", "stream", "params"))


class Service(object):
    """Defines the base concept of a consumed service.
        This is to enable encapsulating the logic of calling a service so usage can be independant of the interface
    """

    __slots__ = ("timeout", "raise_on", "version")

    def __init__(self, version=None, timeout=None, raise_on=(500,), **kwargs):
        self.version = version
        self.timeout = timeout
        self.raise_on = raise_on if type(raise_on) in (tuple, list) else (raise_on,)

    def request(
        self, method, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params
    ):
        """Calls the service at the specified URL using the "CALL" method"""
        raise NotImplementedError("Concrete services must define the request method")

    def get(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "GET" method"""
        return self.request("GET", url=url, headers=headers, timeout=timeout, **params)

    def post(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "POST" method"""
        return self.request("POST", url=url, headers=headers, timeout=timeout, **params)

    def delete(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "DELETE" method"""
        return self.request("DELETE", url=url, headers=headers, timeout=timeout, **params)

    def put(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "PUT" method"""
        return self.request("PUT", url=url, headers=headers, timeout=timeout, **params)

    def trace(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "TRACE" method"""
        return self.request("TRACE", url=url, headers=headers, timeout=timeout, **params)

    def patch(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "PATCH" method"""
        return self.request("PATCH", url=url, headers=headers, timeout=timeout, **params)

    def options(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "OPTIONS" method"""
        return self.request("OPTIONS", url=url, headers=headers, timeout=timeout, **params)

    def head(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "HEAD" method"""
        return self.request("HEAD", url=url, headers=headers, timeout=timeout, **params)

    def connect(self, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params):
        """Calls the service at the specified URL using the "CONNECT" method"""
        return self.request("CONNECT", url=url, headers=headers, timeout=timeout, **params)


class HTTP(Service):
    __slots__ = ("endpoint", "session", "json_transport")

    def __init__(
        self,
        endpoint,
        auth=None,
        version=None,
        headers=empty.dict,
        timeout=None,
        raise_on=(500,),
        json_transport=True,
        **kwargs
    ):
        super().__init__(timeout=timeout, raise_on=raise_on, version=version, **kwargs)
        self.endpoint = endpoint
        self.session = requests.Session()
        self.session.auth = auth
        self.session.headers.update(headers)
        self.json_transport = json_transport

    def request(
        self, method, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params
    ):
        url = "{0}/{1}".format(self.version, url.lstrip("/")) if self.version else url
        kwargs = {"json" if self.json_transport else "params": params}
        response = self.session.request(
            method, self.endpoint + url.format(url_params), headers=headers, **kwargs
        )

        data = BytesIO(response.content)
        content_type, content_params = parse_content_type(response.headers.get("content-type", ""))
        if content_type in input_format:
            data = input_format[content_type](data, **content_params)

        if response.status_code in self.raise_on:
            raise requests.HTTPError(
                "{0} {1} occured for url: {2}".format(response.status_code, response.reason, url)
            )

        return Response(data, response.status_code, response.headers)


class Local(Service):
    __slots__ = ("api", "headers")

    def __init__(
        self, api, version=None, headers=empty.dict, timeout=None, raise_on=(500,), **kwargs
    ):
        super().__init__(timeout=timeout, raise_on=raise_on, version=version, **kwargs)
        self.api = API(api)
        self.headers = headers

    def request(
        self, method, url, url_params=empty.dict, headers=empty.dict, timeout=None, **params
    ):
        function = self.api.http.versioned.get(self.version, {}).get(url, None)
        if not function:
            function = self.api.http.versioned.get(None, {}).get(url, None)

        if not function:
            if 404 in self.raise_on:
                raise requests.HTTPError("404 Not Found occured for url: {0}".format(url))
            return Response("Not Found", 404, {"content-type", "application/json"})

        interface = function.interface.http
        response = falcon.Response()
        request = Request(None, None, empty.dict)
        context = self.api.context_factory(
            api=self.api, api_version=self.version, interface=interface
        )
        interface.set_response_defaults(response)

        params.update(url_params)
        params = interface.gather_parameters(
            request, response, context, api_version=self.version, **params
        )
        errors = interface.validate(params, context)
        if errors:
            interface.render_errors(errors, request, response)
        else:
            interface.render_content(interface.call_function(params), context, request, response)

        data = BytesIO(response.data)
        content_type, content_params = parse_content_type(response._headers.get("content-type", ""))
        if content_type in input_format:
            data = input_format[content_type](data, **content_params)

        status_code = int("".join(re.findall("\d+", response.status)))
        if status_code in self.raise_on:
            raise requests.HTTPError("{0} occured for url: {1}".format(response.status, url))

        return Response(data, status_code, response._headers)


class Socket(Service):
    __slots__ = ("connection_pool", "timeout", "connection", "send_and_receive")

    on_unix = getattr(socket, "AF_UNIX", False)
    Connection = namedtuple("Connection", ("connect_to", "proto", "sockopts"))
    protocols = {
        "tcp": (socket.AF_INET, socket.SOCK_STREAM),
        "udp": (socket.AF_INET, socket.SOCK_DGRAM),
    }
    streams = set(("tcp",))
    datagrams = set(("udp",))
    inet = set(("tcp", "udp"))
    unix = set()

    if on_unix:
        protocols.update(
            {
                "unix_dgram": (socket.AF_UNIX, socket.SOCK_DGRAM),
                "unix_stream": (socket.AF_UNIX, socket.SOCK_STREAM),
            }
        )
        streams.add("unix_stream")
        datagrams.add("unix_dgram")
        unix.update(("unix_stream", "unix_dgram"))

    def __init__(
        self,
        connect_to,
        proto,
        version=None,
        headers=empty.dict,
        timeout=None,
        pool=0,
        raise_on=(500,),
        **kwargs
    ):
        super().__init__(timeout=timeout, raise_on=raise_on, version=version, **kwargs)
        connect_to = tuple(connect_to) if proto in Socket.inet else connect_to
        self.timeout = timeout
        self.connection = Socket.Connection(connect_to, proto, set())
        self.connection_pool = Queue(maxsize=pool if pool else 1)

        if proto in Socket.streams:
            self.send_and_receive = self._stream_send_and_receive
        else:
            self.send_and_receive = self._dgram_send_and_receive

    def settimeout(self, timeout):
        """Set the default timeout"""
        self.timeout = timeout

    def setsockopt(self, *sockopts):
        """Add socket options to set"""
        if type(sockopts[0]) in (list, tuple):
            for sock_opt in sockopts[0]:
                level, option, value = sock_opt
                self.connection.sockopts.add((level, option, value))
        else:
            level, option, value = sockopts
            self.connection.sockopts.add((level, option, value))

    def _register_socket(self):
        """Create/Connect socket, apply options"""
        _socket = socket.socket(*Socket.protocols[self.connection.proto])
        _socket.settimeout(self.timeout)

        # Reconfigure original socket options.
        if self.connection.sockopts:
            for sock_opt in self.connection.sockopts:
                level, option, value = sock_opt
                _socket.setsockopt(level, option, value)

        _socket.connect(self.connection.connect_to)
        return _socket

    def _stream_send_and_receive(self, _socket, message, *args, **kwargs):
        """TCP/Stream sender and receiver"""
        data = BytesIO()

        _socket_fd = _socket.makefile(mode="rwb", encoding="utf-8")
        _socket_fd.write(message.encode("utf-8"))
        _socket_fd.flush()

        for received in _socket_fd:
            data.write(received)
        data.seek(0)

        _socket_fd.close()
        return data

    def _dgram_send_and_receive(self, _socket, message, buffer_size=4096, *args):
        """User Datagram Protocol sender and receiver"""
        _socket.send(message.encode("utf-8"))
        data, address = _socket.recvfrom(buffer_size)
        return BytesIO(data)

    def request(self, message, timeout=False, *args, **kwargs):
        """Populate connection pool, send message, return BytesIO, and cleanup"""
        if not self.connection_pool.full():
            self.connection_pool.put(self._register_socket())

        _socket = self.connection_pool.get()

        # setting timeout to None enables the socket to block.
        if timeout or timeout is None:
            _socket.settimeout(timeout)

        data = self.send_and_receive(_socket, message, *args, **kwargs)

        if self.connection.proto in Socket.streams:
            _socket.shutdown(socket.SHUT_RDWR)

        return Response(data, None, None)



"""hug/api.py

Defines the dynamically generated Hug API object that is responsible for storing all routes and state within a module

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import asyncio
import sys
from collections import OrderedDict, namedtuple
from distutils.util import strtobool
from functools import partial
from itertools import chain
from types import ModuleType
from wsgiref.simple_server import make_server

import falcon
from falcon import HTTP_METHODS

import hug.defaults
import hug.output_format
from hug import introspect
from hug._version import current

INTRO = """
/#######################################################################\\
          `.----``..-------..``.----.
         :/:::::--:---------:--::::://.
        .+::::----##/-/oo+:-##----:::://
        `//::-------/oosoo-------::://.       ##    ##  ##    ##    #####
          .-:------./++o/o-.------::-`   ```  ##    ##  ##    ##  ##
             `----.-./+o+:..----.     `.:///. ########  ##    ## ##
   ```        `----.-::::::------  `.-:::://. ##    ##  ##    ## ##   ####
  ://::--.``` -:``...-----...` `:--::::::-.`  ##    ##  ##   ##   ##    ##
  :/:::::::::-:-     `````      .:::::-.`     ##    ##    ####     ######
   ``.--:::::::.                .:::.`
         ``..::.                .::         EMBRACE THE APIs OF THE FUTURE
             ::-                .:-
             -::`               ::-                   VERSION {0}
             `::-              -::`
              -::-`           -::-
\########################################################################/

 Copyright (C) 2016 Timothy Edmund Crosley
 Under the MIT License

""".format(
    current
)


class InterfaceAPI(object):
    """Defines the per-interface API which defines all shared information for a specific interface, and how it should
        be exposed
    """

    __slots__ = ("api",)

    def __init__(self, api):
        self.api = api


class HTTPInterfaceAPI(InterfaceAPI):
    """Defines the HTTP interface specific API"""

    __slots__ = (
        "routes",
        "versions",
        "base_url",
        "falcon",
        "_output_format",
        "_input_format",
        "versioned",
        "_middleware",
        "_not_found_handlers",
        "sinks",
        "_not_found",
        "_exception_handlers",
    )

    def __init__(self, api, base_url=""):
        super().__init__(api)
        self.versions = set()
        self.routes = OrderedDict()
        self.sinks = OrderedDict()
        self.versioned = OrderedDict()
        self.base_url = base_url

    @property
    def output_format(self):
        return getattr(self, "_output_format", hug.defaults.output_format)

    @output_format.setter
    def output_format(self, formatter):
        self._output_format = formatter

    @property
    def not_found(self):
        """Returns the active not found handler"""
        return getattr(self, "_not_found", self.base_404)

    def urls(self):
        """Returns a generator of all URLs attached to this API"""
        for base_url, mapping in self.routes.items():
            for url, _ in mapping.items():
                yield base_url + url

    def handlers(self):
        """Returns all registered handlers attached to this API"""
        used = []
        for _base_url, mapping in self.routes.items():
            for _url, methods in mapping.items():
                for _method, versions in methods.items():
                    for _version, handler in versions.items():
                        if not handler in used:
                            used.append(handler)
                            yield handler

    def input_format(self, content_type):
        """Returns the set input_format handler for the given content_type"""
        return getattr(self, "_input_format", {}).get(
            content_type, hug.defaults.input_format.get(content_type, None)
        )

    def set_input_format(self, content_type, handler):
        """Sets an input format handler for this Hug API, given the specified content_type"""
        if getattr(self, "_input_format", None) is None:
            self._input_format = {}
        self._input_format[content_type] = handler

    @property
    def middleware(self):
        return getattr(self, "_middleware", None)

    def add_middleware(self, middleware):
        """Adds a middleware object used to process all incoming requests against the API"""
        if self.middleware is None:
            self._middleware = []
        self.middleware.append(middleware)

    def add_sink(self, sink, url, base_url=""):
        base_url = base_url or self.base_url
        self.sinks.setdefault(base_url, OrderedDict())
        self.sinks[base_url][url] = sink

    def exception_handlers(self, version=None):
        if not hasattr(self, "_exception_handlers"):
            return None

        return self._exception_handlers.get(version, self._exception_handlers.get(None, None))

    def add_exception_handler(self, exception_type, error_handler, versions=(None,)):
        """Adds a error handler to the hug api"""
        versions = (versions,) if not isinstance(versions, (tuple, list)) else versions
        if not hasattr(self, "_exception_handlers"):
            self._exception_handlers = {}

        for version in versions:
            placement = self._exception_handlers.setdefault(version, OrderedDict())
            placement[exception_type] = (error_handler,) + placement.get(exception_type, tuple())

    def extend(self, http_api, route="", base_url="", **kwargs):
        """Adds handlers from a different Hug API to this one - to create a single API"""
        self.versions.update(http_api.versions)
        base_url = base_url or self.base_url

        for _router_base_url, routes in http_api.routes.items():
            self.routes.setdefault(base_url, OrderedDict())
            for item_route, handler in routes.items():
                for _method, versions in handler.items():
                    for _version, function in versions.items():
                        function.interface.api = self.api
                self.routes[base_url].setdefault(route + item_route, {}).update(handler)

        for _sink_base_url, sinks in http_api.sinks.items():
            for url, sink in sinks.items():
                self.add_sink(sink, route + url, base_url=base_url)

        for middleware in http_api.middleware or ():
            self.add_middleware(middleware)

        for version, handler in getattr(http_api, "_exception_handlers", {}).items():
            for exception_type, exception_handlers in handler.items():
                target_exception_handlers = self.exception_handlers(version) or {}
                for exception_handler in exception_handlers:
                    if exception_type not in target_exception_handlers:
                        self.add_exception_handler(exception_type, exception_handler, version)

        for input_format, input_format_handler in getattr(http_api, "_input_format", {}).items():
            if not input_format in getattr(self, "_input_format", {}):
                self.set_input_format(input_format, input_format_handler)

        for version, handler in http_api.not_found_handlers.items():
            if version not in self.not_found_handlers:
                self.set_not_found_handler(handler, version)

    @property
    def not_found_handlers(self):
        return getattr(self, "_not_found_handlers", {})

    def set_not_found_handler(self, handler, version=None):
        """Sets the not_found handler for the specified version of the api"""
        if not self.not_found_handlers:
            self._not_found_handlers = {}

        self.not_found_handlers[version] = handler

    def documentation(self, base_url=None, api_version=None, prefix=""):
        """Generates and returns documentation for this API endpoint"""
        documentation = OrderedDict()
        base_url = self.base_url if base_url is None else base_url
        overview = self.api.doc
        if overview:
            documentation["overview"] = overview

        version_dict = OrderedDict()
        versions = self.versions
        versions_list = list(versions)
        if None in versions_list:
            versions_list.remove(None)
        if False in versions_list:
            versions_list.remove(False)
        if api_version is None and len(versions_list) > 0:
            api_version = max(versions_list)
            documentation["version"] = api_version
        elif api_version is not None:
            documentation["version"] = api_version
        if versions_list:
            documentation["versions"] = versions_list
        for router_base_url, routes in self.routes.items():
            for url, methods in routes.items():
                for method, method_versions in methods.items():
                    for version, handler in method_versions.items():
                        if getattr(handler, "private", False):
                            continue
                        if version is None:
                            applies_to = versions
                        else:
                            applies_to = (version,)
                        for version in applies_to:
                            if api_version and version != api_version:
                                continue
                            if base_url and router_base_url != base_url:
                                continue
                            doc = version_dict.setdefault(url, OrderedDict())
                            doc[method] = handler.documentation(
                                doc.get(method, None),
                                version=version,
                                prefix=prefix,
                                base_url=router_base_url,
                                url=url,
                            )
        documentation["handlers"] = version_dict
        return documentation

    def serve(self, host="", port=8000, no_documentation=False, display_intro=True):
        """Runs the basic hug development server against this API"""
        if no_documentation:
            api = self.server(None)
        else:
            api = self.server()

        if display_intro:
            print(INTRO)

        httpd = make_server(host, port, api)
        print("Serving on {0}:{1}...".format(host, port))
        httpd.serve_forever()

    @staticmethod
    def base_404(request, response, *args, **kwargs):
        """Defines the base 404 handler"""
        response.status = falcon.HTTP_NOT_FOUND

    def determine_version(self, request, api_version=None):
        """Determines the appropriate version given the set api_version, the request header, and URL query params"""
        if api_version is False:
            api_version = None
            for version in self.versions:
                if version and "v{0}".format(version) in request.path:
                    api_version = version
                    break

        request_version = set()
        if api_version is not None:
            request_version.add(api_version)

        version_header = request.get_header("X-API-VERSION")
        if version_header:
            request_version.add(version_header)

        version_param = request.get_param("api_version")
        if version_param is not None:
            request_version.add(version_param)

        if len(request_version) > 1:
            raise ValueError("You are requesting conflicting versions")

        return next(iter(request_version or (None,)))

    def documentation_404(self, base_url=None):
        """Returns a smart 404 page that contains documentation for the written API"""
        base_url = self.base_url if base_url is None else base_url

        def handle_404(request, response, *args, **kwargs):
            url_prefix = request.forwarded_uri[:-1]
            if request.path and request.path != "/":
                url_prefix = request.forwarded_uri.split(request.path)[0]

            to_return = OrderedDict()
            to_return["404"] = (
                "The API call you tried to make was not defined. "
                "Here's a definition of the API to help you get going :)"
            )
            to_return["documentation"] = self.documentation(
                base_url, self.determine_version(request, False), prefix=url_prefix
            )

            if self.output_format == hug.output_format.json:
                response.data = hug.output_format.json(to_return, indent=4, separators=(",", ": "))
                response.content_type = "application/json; charset=utf-8"
            else:
                response.data = self.output_format(to_return, request=request, response=response)
                response.content_type = self.output_format.content_type

            response.status = falcon.HTTP_NOT_FOUND

        handle_404.interface = True
        return handle_404

    def version_router(
        self, request, response, api_version=None, versions=None, not_found=None, **kwargs
    ):
        """Intelligently routes a request to the correct handler based on the version being requested"""
        if versions is None:
            versions = {}
        request_version = self.determine_version(request, api_version)
        if request_version:
            request_version = int(request_version)
        versions.get(request_version or False, versions.get(None, not_found))(
            request, response, api_version=api_version, **kwargs
        )

    def server(self, default_not_found=True, base_url=None):
        """Returns a WSGI compatible API server for the given Hug API module"""
        falcon_api = self.falcon = falcon.API(middleware=self.middleware)
        if not self.api.future:
            falcon_api.req_options.keep_blank_qs_values = False
            falcon_api.req_options.auto_parse_qs_csv = True
            falcon_api.req_options.strip_url_path_trailing_slash = True

        default_not_found = self.documentation_404() if default_not_found is True else None
        base_url = self.base_url if base_url is None else base_url

        not_found_handler = default_not_found
        self.api._ensure_started()
        if self.not_found_handlers:
            if len(self.not_found_handlers) == 1 and None in self.not_found_handlers:
                not_found_handler = self.not_found_handlers[None]
            else:
                not_found_handler = partial(
                    self.version_router,
                    api_version=False,
                    versions=self.not_found_handlers,
                    not_found=default_not_found,
                )
                not_found_handler.interface = True

        if not_found_handler:
            falcon_api.add_sink(not_found_handler)
            self._not_found = not_found_handler

        for sink_base_url, sinks in self.sinks.items():
            for url, extra_sink in sinks.items():
                falcon_api.add_sink(extra_sink, sink_base_url + url + "(?P<path>.*)")

        for router_base_url, routes in self.routes.items():
            for url, methods in routes.items():
                router = {}
                for method, versions in methods.items():
                    method_function = "on_{0}".format(method.lower())
                    if len(versions) == 1 and None in versions.keys():
                        router[method_function] = versions[None]
                    else:
                        router[method_function] = partial(
                            self.version_router, versions=versions, not_found=not_found_handler
                        )

                router = namedtuple("Router", router.keys())(**router)
                falcon_api.add_route(router_base_url + url, router)
                if self.versions and self.versions != (None,):
                    falcon_api.add_route(router_base_url + "/v{api_version}" + url, router)

        def error_serializer(request, response, error):
            response.content_type = self.output_format.content_type
            response.body = self.output_format(
                {"errors": {error.title: error.description}}, request, response
            )

        falcon_api.set_error_serializer(error_serializer)
        return falcon_api


HTTPInterfaceAPI.base_404.interface = True


class CLIInterfaceAPI(InterfaceAPI):
    """Defines the CLI interface specific API"""

    __slots__ = ("commands", "error_exit_codes", "_output_format")

    def __init__(self, api, version="", error_exit_codes=False):
        super().__init__(api)
        self.commands = {}
        self.error_exit_codes = error_exit_codes

    def __call__(self, args=None):
        """Routes to the correct command line tool"""
        self.api._ensure_started()
        args = sys.argv if args is None else args
        if not len(args) > 1 or not args[1] in self.commands:
            print(str(self))
            return sys.exit(1)

        command = args.pop(1)
        result = self.commands.get(command)()

        if self.error_exit_codes and bool(strtobool(result.decode("utf-8"))) is False:
            sys.exit(1)

    def handlers(self):
        """Returns all registered handlers attached to this API"""
        return self.commands.values()

    def extend(self, cli_api, command_prefix="", sub_command="", **kwargs):
        """Extends this CLI api with the commands present in the provided cli_api object"""
        if sub_command and command_prefix:
            raise ValueError(
                "It is not currently supported to provide both a command_prefix and sub_command"
            )

        if sub_command:
            self.commands[sub_command] = cli_api
        else:
            for name, command in cli_api.commands.items():
                self.commands["{}{}".format(command_prefix, name)] = command

    @property
    def output_format(self):
        return getattr(self, "_output_format", hug.defaults.cli_output_format)

    @output_format.setter
    def output_format(self, formatter):
        self._output_format = formatter

    def __str__(self):
        return "{0}\n\nAvailable Commands:{1}\n".format(
            self.api.doc or self.api.name, "\n\n\t- " + "\n\t- ".join(self.commands.keys())
        )


class ModuleSingleton(type):
    """Defines the module level __hug__ singleton"""

    def __call__(cls, module=None, *args, **kwargs):
        if isinstance(module, API):
            return module

        if type(module) == str:
            if module not in sys.modules:
                sys.modules[module] = ModuleType(module)
            module = sys.modules[module]
        elif module is None:
            return super().__call__(*args, **kwargs)

        if not "__hug__" in module.__dict__:

            def api_auto_instantiate(*args, **kwargs):
                if not hasattr(module, "__hug_serving__"):
                    module.__hug_wsgi__ = module.__hug__.http.server()
                    module.__hug_serving__ = True
                return module.__hug_wsgi__(*args, **kwargs)

            module.__hug__ = super().__call__(module, *args, **kwargs)
            module.__hug_wsgi__ = api_auto_instantiate
        return module.__hug__


class API(object, metaclass=ModuleSingleton):
    """Stores the information necessary to expose API calls within this module externally"""

    __slots__ = (
        "module",
        "_directives",
        "_http",
        "_cli",
        "_context",
        "_context_factory",
        "_delete_context",
        "_startup_handlers",
        "started",
        "name",
        "doc",
        "future",
        "cli_error_exit_codes",
    )

    def __init__(self, module=None, name="", doc="", cli_error_exit_codes=False, future=False):
        self.module = module
        if module:
            self.name = name or module.__name__ or ""
            self.doc = doc or module.__doc__ or ""
        else:
            self.name = name
            self.doc = doc
        self.started = False
        self.cli_error_exit_codes = cli_error_exit_codes
        self.future = future

    def directives(self):
        """Returns all directives applicable to this Hug API"""
        directive_sources = chain(
            hug.defaults.directives.items(), getattr(self, "_directives", {}).items()
        )
        return {
            "hug_" + directive_name: directive for directive_name, directive in directive_sources
        }

    def directive(self, name, default=None):
        """Returns the loaded directive with the specified name, or default if passed name is not present"""
        return getattr(self, "_directives", {}).get(
            name, hug.defaults.directives.get(name, default)
        )

    def add_directive(self, directive):
        self._directives = getattr(self, "_directives", {})
        self._directives[directive.__name__] = directive

    def handlers(self):
        """Returns all registered handlers attached to this API"""
        if getattr(self, "_http", None):
            yield from self.http.handlers()
        if getattr(self, "_cli", None):
            yield from self.cli.handlers()

    @property
    def http(self):
        if not hasattr(self, "_http"):
            self._http = HTTPInterfaceAPI(self)
        return self._http

    @property
    def cli(self):
        if not hasattr(self, "_cli"):
            self._cli = CLIInterfaceAPI(self, error_exit_codes=self.cli_error_exit_codes)
        return self._cli

    @property
    def context_factory(self):
        return getattr(self, "_context_factory", hug.defaults.context_factory)

    @context_factory.setter
    def context_factory(self, context_factory_):
        self._context_factory = context_factory_

    @property
    def delete_context(self):
        return getattr(self, "_delete_context", hug.defaults.delete_context)

    @delete_context.setter
    def delete_context(self, delete_context_):
        self._delete_context = delete_context_

    @property
    def context(self):
        if not hasattr(self, "_context"):
            self._context = {}
        return self._context

    def extend(self, api, route="", base_url="", http=True, cli=True, **kwargs):
        """Adds handlers from a different Hug API to this one - to create a single API"""
        api = API(api)

        if http and hasattr(api, "_http"):
            self.http.extend(api.http, route, base_url, **kwargs)

        if cli and hasattr(api, "_cli"):
            self.cli.extend(api.cli, **kwargs)

        for directive in getattr(api, "_directives", {}).values():
            self.add_directive(directive)

        for startup_handler in api.startup_handlers or ():
            self.add_startup_handler(startup_handler)

    def add_startup_handler(self, handler):
        """Adds a startup handler to the hug api"""
        if not self.startup_handlers:
            self._startup_handlers = []

        self.startup_handlers.append(handler)

    def _ensure_started(self):
        """Marks the API as started and runs all startup handlers"""
        if not self.started:
            async_handlers = [
                startup_handler
                for startup_handler in self.startup_handlers
                if introspect.is_coroutine(startup_handler)
            ]
            if async_handlers:
                loop = asyncio.get_event_loop()
                loop.run_until_complete(
                    asyncio.gather(*[handler(self) for handler in async_handlers], loop=loop)
                )
            for startup_handler in self.startup_handlers:
                if not startup_handler in async_handlers:
                    startup_handler(self)

    @property
    def startup_handlers(self):
        return getattr(self, "_startup_handlers", ())


def from_object(obj):
    """Returns a Hug API instance from a given object (function, class, instance)"""
    return API(obj.__module__)



"""hug/interface.py

Defines the various interface hug provides to expose routes to functions

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import argparse
import asyncio
import os
import sys
from collections import OrderedDict
from functools import lru_cache, partial, wraps

import falcon
from falcon import HTTP_BAD_REQUEST

import hug._empty as empty
import hug.api
import hug.output_format
import hug.types as types
from hug import introspect
from hug.exceptions import InvalidTypeData
from hug.format import parse_content_type
from hug.types import (
    MarshmallowInputSchema,
    MarshmallowReturnSchema,
    Multiple,
    OneOf,
    SmartBoolean,
    Text,
    text,
)


def asyncio_call(function, *args, **kwargs):
    loop = asyncio.get_event_loop()
    if loop.is_running():
        return function(*args, **kwargs)

    function = asyncio.ensure_future(function(*args, **kwargs), loop=loop)
    loop.run_until_complete(function)
    return function.result()


class Interfaces(object):
    """Defines the per-function singleton applied to hugged functions defining common data needed by all interfaces"""

    def __init__(self, function, args=None):
        self.api = hug.api.from_object(function)
        self.spec = getattr(function, "original", function)
        self.arguments = introspect.arguments(function)
        self.name = introspect.name(function)
        self._function = function

        self.is_coroutine = introspect.is_coroutine(self.spec)
        if self.is_coroutine:
            self.spec = getattr(self.spec, "__wrapped__", self.spec)

        self.takes_args = introspect.takes_args(self.spec)
        self.takes_kwargs = introspect.takes_kwargs(self.spec)

        self.parameters = list(introspect.arguments(self.spec, self.takes_kwargs + self.takes_args))
        if self.takes_kwargs:
            self.kwarg = self.parameters.pop(-1)
        if self.takes_args:
            self.arg = self.parameters.pop(-1)
        self.parameters = tuple(self.parameters)
        self.defaults = dict(zip(reversed(self.parameters), reversed(self.spec.__defaults__ or ())))
        self.required = self.parameters[: -(len(self.spec.__defaults__ or ())) or None]
        self.is_method = introspect.is_method(self.spec) or introspect.is_method(function)
        if self.is_method:
            self.required = self.required[1:]
            self.parameters = self.parameters[1:]

        self.all_parameters = set(self.parameters)
        if self.spec is not function:
            self.all_parameters.update(self.arguments)

        if args is not None:
            transformers = args
        else:
            transformers = self.spec.__annotations__

        self.transform = transformers.get("return", None)
        self.directives = {}
        self.input_transformations = {}
        for name, transformer in transformers.items():
            if isinstance(transformer, str):
                continue
            elif hasattr(transformer, "directive"):
                self.directives[name] = transformer
                continue

            if hasattr(transformer, "from_string"):
                transformer = transformer.from_string
            elif hasattr(transformer, "load"):
                transformer = MarshmallowInputSchema(transformer)
            elif hasattr(transformer, "deserialize"):
                transformer = transformer.deserialize

            self.input_transformations[name] = transformer

    def __call__(__hug_internal_self, *args, **kwargs):  # noqa: N805
        """"Calls the wrapped function, uses __hug_internal_self incase self is passed in as a kwarg from the wrapper"""
        if not __hug_internal_self.is_coroutine:
            return __hug_internal_self._function(*args, **kwargs)

        return asyncio_call(__hug_internal_self._function, *args, **kwargs)


class Interface(object):
    """Defines the basic hug interface object, which is responsible for wrapping a user defined function and providing
       all the info requested in the function as well as the route

       A Interface object should be created for every kind of protocal hug supports
    """

    __slots__ = (
        "interface",
        "_api",
        "defaults",
        "parameters",
        "required",
        "_outputs",
        "on_invalid",
        "requires",
        "validate_function",
        "transform",
        "examples",
        "output_doc",
        "wrapped",
        "directives",
        "all_parameters",
        "raise_on_invalid",
        "invalid_outputs",
        "map_params",
        "input_transformations",
    )

    def __init__(self, route, function):
        if route.get("api", None):
            self._api = route["api"]
        if "examples" in route:
            self.examples = route["examples"]
        function_args = route.get("args")
        if not hasattr(function, "interface"):
            function.__dict__["interface"] = Interfaces(function, function_args)

        self.interface = function.interface
        self.requires = route.get("requires", ())
        if "validate" in route:
            self.validate_function = route["validate"]
        if "output_invalid" in route:
            self.invalid_outputs = route["output_invalid"]

        if not "parameters" in route:
            self.defaults = self.interface.defaults
            self.parameters = self.interface.parameters
            self.all_parameters = self.interface.all_parameters
            self.required = self.interface.required
        else:
            self.defaults = route.get("defaults", {})
            self.parameters = tuple(route["parameters"])
            self.all_parameters = set(route["parameters"])
            self.required = tuple(
                [parameter for parameter in self.parameters if parameter not in self.defaults]
            )

        if "map_params" in route:
            self.map_params = route["map_params"]
            for interface_name, internal_name in self.map_params.items():
                if internal_name in self.defaults:
                    self.defaults[interface_name] = self.defaults.pop(internal_name)
                if internal_name in self.parameters:
                    self.parameters = [
                        interface_name if param == internal_name else param
                        for param in self.parameters
                    ]
                if internal_name in self.all_parameters:
                    self.all_parameters.remove(internal_name)
                    self.all_parameters.add(interface_name)
                if internal_name in self.required:
                    self.required = tuple(
                        [
                            interface_name if param == internal_name else param
                            for param in self.required
                        ]
                    )

            reverse_mapping = {
                internal: interface for interface, internal in self.map_params.items()
            }
            self.input_transformations = {
                reverse_mapping.get(name, name): transform
                for name, transform in self.interface.input_transformations.items()
            }
        else:
            self.input_transformations = self.interface.input_transformations

        if "output" in route:
            self.outputs = route["output"]

        self.transform = route.get("transform", None)
        if self.transform is None and not isinstance(self.interface.transform, (str, type(None))):
            self.transform = self.interface.transform

        if hasattr(self.transform, "dump"):
            self.transform = MarshmallowReturnSchema(self.transform)
            self.output_doc = self.transform.__doc__
        elif self.transform or self.interface.transform:
            output_doc = self.transform or self.interface.transform
            self.output_doc = output_doc if type(output_doc) is str else output_doc.__doc__

        self.raise_on_invalid = route.get("raise_on_invalid", False)
        if "on_invalid" in route:
            self.on_invalid = route["on_invalid"]
        elif self.transform:
            self.on_invalid = self.transform

        defined_directives = self.api.directives()
        used_directives = set(self.parameters).intersection(defined_directives)
        self.directives = {
            directive_name: defined_directives[directive_name] for directive_name in used_directives
        }
        self.directives.update(self.interface.directives)

    @property
    def api(self):
        return getattr(self, "_api", self.interface.api)

    @property
    def outputs(self):
        return getattr(self, "_outputs", None)

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs  # pragma: no cover - generally re-implemented by sub classes

    def validate(self, input_parameters, context):
        """Runs all set type transformers / validators against the provided input parameters and returns any errors"""
        errors = {}

        for key, type_handler in self.input_transformations.items():
            if self.raise_on_invalid:
                if key in input_parameters:
                    input_parameters[key] = self.initialize_handler(
                        type_handler, input_parameters[key], context=context
                    )
            else:
                try:
                    if key in input_parameters:
                        input_parameters[key] = self.initialize_handler(
                            type_handler, input_parameters[key], context=context
                        )
                except InvalidTypeData as error:
                    errors[key] = error.reasons or str(error)
                except Exception as error:
                    if hasattr(error, "args") and error.args:
                        errors[key] = error.args[0]
                    else:
                        errors[key] = str(error)
        for require in self.required:
            if not require in input_parameters:
                errors[require] = "Required parameter '{}' not supplied".format(require)
        if not errors and getattr(self, "validate_function", False):
            errors = self.validate_function(input_parameters)
        return errors

    def check_requirements(self, request=None, response=None, context=None):
        """Checks to see if all requirements set pass

           if all requirements pass nothing will be returned
           otherwise, the error reported will be returned
        """
        for requirement in self.requires:
            conclusion = requirement(
                response=response, request=request, context=context, module=self.api.module
            )
            if conclusion and conclusion is not True:
                return conclusion

    def documentation(self, add_to=None):
        """Produces general documentation for the interface"""
        doc = OrderedDict if add_to is None else add_to

        usage = self.interface.spec.__doc__
        if usage:
            doc["usage"] = usage
        if getattr(self, "requires", None):
            doc["requires"] = [
                getattr(requirement, "__doc__", requirement.__name__)
                for requirement in self.requires
            ]
        doc["outputs"] = OrderedDict()
        doc["outputs"]["format"] = self.outputs.__doc__
        doc["outputs"]["content_type"] = self.outputs.content_type
        parameters = [
            param
            for param in self.parameters
            if not param in ("request", "response", "self")
            and not param in ("api_version", "body")
            and not param.startswith("hug_")
            and not hasattr(param, "directive")
        ]
        if parameters:
            inputs = doc.setdefault("inputs", OrderedDict())
            types = self.interface.spec.__annotations__
            for argument in parameters:
                kind = types.get(argument, text)
                if getattr(kind, "directive", None) is True:
                    continue

                input_definition = inputs.setdefault(argument, OrderedDict())
                input_definition["type"] = kind if isinstance(kind, str) else kind.__doc__
                default = self.defaults.get(argument, None)
                if default is not None:
                    input_definition["default"] = default

        return doc

    def _rewrite_params(self, params):
        for interface_name, internal_name in self.map_params.items():
            if interface_name in params:
                params[internal_name] = params.pop(interface_name)

    @staticmethod
    def cleanup_parameters(parameters, exception=None):
        for _parameter, directive in parameters.items():
            if hasattr(directive, "cleanup"):
                directive.cleanup(exception=exception)

    @staticmethod
    def initialize_handler(handler, value, context):
        try:  # It's easier to ask for forgiveness than for permission
            return handler(value, context=context)
        except TypeError:
            return handler(value)


class Local(Interface):
    """Defines the Interface responsible for exposing functions locally"""

    __slots__ = ("skip_directives", "skip_validation", "version")

    def __init__(self, route, function):
        super().__init__(route, function)
        self.version = route.get("version", None)
        if "skip_directives" in route:
            self.skip_directives = True
        if "skip_validation" in route:
            self.skip_validation = True

        self.interface.local = self

    def __get__(self, instance, kind):
        """Support instance methods"""
        return partial(self.__call__, instance) if instance else self.__call__

    @property
    def __name__(self):
        return self.interface.spec.__name__

    @property
    def __module__(self):
        return self.interface.spec.__module__

    def __call__(self, *args, **kwargs):
        context = self.api.context_factory(api=self.api, api_version=self.version, interface=self)
        """Defines how calling the function locally should be handled"""

        for _requirement in self.requires:
            lacks_requirement = self.check_requirements(context=context)
            if lacks_requirement:
                self.api.delete_context(context, lacks_requirement=lacks_requirement)
                return self.outputs(lacks_requirement) if self.outputs else lacks_requirement

        for index, argument in enumerate(args):
            kwargs[self.parameters[index]] = argument

        if not getattr(self, "skip_directives", False):
            for parameter, directive in self.directives.items():
                if parameter in kwargs:
                    continue
                arguments = (self.defaults[parameter],) if parameter in self.defaults else ()
                kwargs[parameter] = directive(
                    *arguments,
                    api=self.api,
                    api_version=self.version,
                    interface=self,
                    context=context
                )

        if not getattr(self, "skip_validation", False):
            errors = self.validate(kwargs, context)
            if errors:
                errors = {"errors": errors}
                if getattr(self, "on_invalid", False):
                    errors = self.on_invalid(errors)
                outputs = getattr(self, "invalid_outputs", self.outputs)
                self.api.delete_context(context, errors=errors)
                return outputs(errors) if outputs else errors

        if getattr(self, "map_params", None):
            self._rewrite_params(kwargs)
        try:
            result = self.interface(**kwargs)
            if self.transform:
                if hasattr(self.transform, "context"):
                    self.transform.context = context
                result = self.transform(result)
        except Exception as exception:
            self.cleanup_parameters(kwargs, exception=exception)
            self.api.delete_context(context, exception=exception)
            raise exception
        self.cleanup_parameters(kwargs)
        self.api.delete_context(context)
        return self.outputs(result) if self.outputs else result


class CLI(Interface):
    """Defines the Interface responsible for exposing functions to the CLI"""

    def __init__(self, route, function):
        super().__init__(route, function)
        if not self.outputs:
            self.outputs = self.api.cli.output_format

        self.interface.cli = self
        self.reaffirm_types = {}
        use_parameters = list(self.interface.parameters)
        self.additional_options = getattr(
            self.interface, "arg", getattr(self.interface, "kwarg", False)
        )
        if self.additional_options:
            use_parameters.append(self.additional_options)

        used_options = {"h", "help"}
        nargs_set = self.interface.takes_args or self.interface.takes_kwargs

        class CustomArgumentParser(argparse.ArgumentParser):
            exit_callback = None

            def exit(self, status=0, message=None):
                if self.exit_callback:
                    self.exit_callback(message)
                super().exit(status, message)

        self.parser = CustomArgumentParser(
            description=route.get("doc", self.interface.spec.__doc__)
        )
        if "version" in route:
            self.parser.add_argument(
                "-v",
                "--version",
                action="version",
                version="{0} {1}".format(
                    route.get("name", self.interface.spec.__name__), route["version"]
                ),
            )
            used_options.update(("v", "version"))

        self.context_tranforms = []
        for option in use_parameters:

            if option in self.directives:
                continue

            if option in self.interface.required or option == self.additional_options:
                args = (option,)
            else:
                short_option = option[0]
                while short_option in used_options and len(short_option) < len(option):
                    short_option = option[: len(short_option) + 1]

                used_options.add(short_option)
                used_options.add(option)
                if short_option != option:
                    args = ("-{0}".format(short_option), "--{0}".format(option))
                else:
                    args = ("--{0}".format(option),)

            kwargs = {}
            if option in self.defaults:
                kwargs["default"] = self.defaults[option]
            if option in self.interface.input_transformations:
                transform = self.interface.input_transformations[option]
                kwargs["type"] = transform
                kwargs["help"] = transform.__doc__
                if transform in (list, tuple) or isinstance(transform, types.Multiple):
                    kwargs["action"] = "append"
                    kwargs["type"] = Text()
                    self.reaffirm_types[option] = transform
                elif transform == bool or isinstance(transform, type(types.boolean)):
                    kwargs["action"] = "store_true"
                    self.reaffirm_types[option] = transform
                elif isinstance(transform, types.OneOf):
                    kwargs["choices"] = transform.values
            elif (
                option in self.interface.spec.__annotations__
                and type(self.interface.spec.__annotations__[option]) == str
            ):
                kwargs["help"] = option
            if (
                kwargs.get("type", None) == bool or kwargs.get("action", None) == "store_true"
            ) and not kwargs["default"]:
                kwargs["action"] = "store_true"
                kwargs.pop("type", None)
            elif kwargs.get("action", None) == "store_true":
                kwargs.pop("action", None) == "store_true"

            if option == self.additional_options:
                kwargs["nargs"] = "*"
            elif (
                not nargs_set
                and kwargs.get("action", None) == "append"
                and not option in self.interface.defaults
            ):
                kwargs["nargs"] = "*"
                kwargs.pop("action", "")
                nargs_set = True

            self.parser.add_argument(*args, **kwargs)

        self.api.cli.commands[route.get("name", self.interface.spec.__name__)] = self

    def output(self, data, context):
        """Outputs the provided data using the transformations and output format specified for this CLI endpoint"""
        if self.transform:
            if hasattr(self.transform, "context"):
                self.transform.context = context
            data = self.transform(data)
        if hasattr(data, "read"):
            data = data.read().decode("utf8")
        if data is not None:
            data = self.outputs(data)
            if data:
                sys.stdout.buffer.write(data)
                if not data.endswith(b"\n"):
                    sys.stdout.buffer.write(b"\n")
        return data

    def __call__(self):
        """Calls the wrapped function through the lens of a CLI ran command"""
        context = self.api.context_factory(api=self.api, argparse=self.parser, interface=self)

        def exit_callback(message):
            self.api.delete_context(context, errors=message)

        self.parser.exit_callback = exit_callback

        self.api._ensure_started()
        for requirement in self.requires:
            conclusion = requirement(request=sys.argv, module=self.api.module, context=context)
            if conclusion and conclusion is not True:
                self.api.delete_context(context, lacks_requirement=conclusion)
                return self.output(conclusion, context)

        if self.interface.is_method:
            self.parser.prog = "%s %s" % (self.api.module.__name__, self.interface.name)

        known, unknown = self.parser.parse_known_args()
        pass_to_function = vars(known)
        for option, directive in self.directives.items():
            arguments = (self.defaults[option],) if option in self.defaults else ()
            pass_to_function[option] = directive(
                *arguments, api=self.api, argparse=self.parser, context=context, interface=self
            )

        for field, type_handler in self.reaffirm_types.items():
            if field in pass_to_function:
                pass_to_function[field] = self.initialize_handler(
                    type_handler, pass_to_function[field], context=context
                )

        if getattr(self, "validate_function", False):
            errors = self.validate_function(pass_to_function)
            if errors:
                self.api.delete_context(context, errors=errors)
                return self.output(errors, context)

        args = None
        if self.additional_options:
            args = []
            for parameter in self.interface.parameters:
                if parameter in pass_to_function:
                    args.append(pass_to_function.pop(parameter))
            args.extend(pass_to_function.pop(self.additional_options, ()))
            if self.interface.takes_kwargs:
                add_options_to = None
                for option in unknown:
                    if option.startswith("--"):
                        if add_options_to:
                            value = pass_to_function[add_options_to]
                            if len(value) == 1:
                                pass_to_function[add_options_to] = value[0]
                            elif value == []:
                                pass_to_function[add_options_to] = True
                        add_options_to = option[2:]
                        pass_to_function.setdefault(add_options_to, [])
                    elif add_options_to:
                        pass_to_function[add_options_to].append(option)

        if getattr(self, "map_params", None):
            self._rewrite_params(pass_to_function)

        try:
            if args:
                result = self.output(self.interface(*args, **pass_to_function), context)
            else:
                result = self.output(self.interface(**pass_to_function), context)
        except Exception as exception:
            self.cleanup_parameters(pass_to_function, exception=exception)
            self.api.delete_context(context, exception=exception)
            raise exception
        self.cleanup_parameters(pass_to_function)
        self.api.delete_context(context)
        return result


class HTTP(Interface):
    """Defines the interface responsible for wrapping functions and exposing them via HTTP based on the route"""

    __slots__ = (
        "_params_for_outputs_state",
        "_params_for_invalid_outputs_state",
        "_params_for_transform_state",
        "_params_for_on_invalid",
        "set_status",
        "response_headers",
        "transform",
        "input_transformations",
        "examples",
        "wrapped",
        "catch_exceptions",
        "parse_body",
        "private",
        "on_invalid",
        "inputs",
    )
    AUTO_INCLUDE = {"request", "response"}

    def __init__(self, route, function, catch_exceptions=True):
        super().__init__(route, function)
        self.catch_exceptions = catch_exceptions
        self.parse_body = "parse_body" in route
        self.set_status = route.get("status", False)
        self.response_headers = tuple(route.get("response_headers", {}).items())
        self.private = "private" in route
        self.inputs = route.get("inputs", {})

        if "on_invalid" in route:
            self._params_for_on_invalid = introspect.takes_arguments(
                self.on_invalid, *self.AUTO_INCLUDE
            )
        elif self.transform:
            self._params_for_on_invalid = self._params_for_transform

        self.api.http.versions.update(route.get("versions", (None,)))

        self.interface.http = self

    @property
    def _params_for_outputs(self):
        if not hasattr(self, "_params_for_outputs_state"):
            self._params_for_outputs_state = introspect.takes_arguments(
                self.outputs, *self.AUTO_INCLUDE
            )
        return self._params_for_outputs_state

    @property
    def _params_for_invalid_outputs(self):
        if not hasattr(self, "_params_for_invalid_outputs_state"):
            self._params_for_invalid_outputs_state = introspect.takes_arguments(
                self.invalid_outputs, *self.AUTO_INCLUDE
            )
        return self._params_for_invalid_outputs_state

    @property
    def _params_for_transform(self):
        if not hasattr(self, "_params_for_transform_state"):
            self._params_for_transform_state = introspect.takes_arguments(
                self.transform, *self.AUTO_INCLUDE
            )
        return self._params_for_transform_state

    def gather_parameters(self, request, response, context, api_version=None, **input_parameters):
        """Gathers and returns all parameters that will be used for this endpoint"""
        input_parameters.update(request.params)

        if self.parse_body and request.content_length:
            body = request.bounded_stream
            content_type, content_params = parse_content_type(request.content_type)
            body_formatter = body and self.inputs.get(
                content_type, self.api.http.input_format(content_type)
            )
            if body_formatter:
                body = body_formatter(body, content_length=request.content_length, **content_params)
            if "body" in self.all_parameters:
                input_parameters["body"] = body
            if isinstance(body, dict):
                input_parameters.update(body)
        elif "body" in self.all_parameters:
            input_parameters["body"] = None

        if "request" in self.all_parameters:
            input_parameters["request"] = request
        if "response" in self.all_parameters:
            input_parameters["response"] = response
        if "api_version" in self.all_parameters:
            input_parameters["api_version"] = api_version
        for parameter, directive in self.directives.items():
            arguments = (self.defaults[parameter],) if parameter in self.defaults else ()
            input_parameters[parameter] = directive(
                *arguments,
                response=response,
                request=request,
                api=self.api,
                api_version=api_version,
                context=context,
                interface=self
            )
        return input_parameters

    @property
    def outputs(self):
        return getattr(self, "_outputs", self.api.http.output_format)

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    def transform_data(self, data, request=None, response=None, context=None):
        transform = self.transform
        if hasattr(transform, "context"):
            self.transform.context = context
        """Runs the transforms specified on this endpoint with the provided data, returning the data modified"""
        if transform and not (isinstance(transform, type) and isinstance(data, transform)):
            if self._params_for_transform:
                return transform(
                    data, **self._arguments(self._params_for_transform, request, response)
                )
            else:
                return transform(data)
        return data

    def content_type(self, request=None, response=None):
        """Returns the content type that should be used by default for this endpoint"""
        if callable(self.outputs.content_type):
            return self.outputs.content_type(request=request, response=response)
        else:
            return self.outputs.content_type

    def invalid_content_type(self, request=None, response=None):
        """Returns the content type that should be used by default on validation errors"""
        if callable(self.invalid_outputs.content_type):
            return self.invalid_outputs.content_type(request=request, response=response)
        else:
            return self.invalid_outputs.content_type

    def _arguments(self, requested_params, request=None, response=None):
        if requested_params:
            arguments = {}
            if "response" in requested_params:
                arguments["response"] = response
            if "request" in requested_params:
                arguments["request"] = request
            return arguments

        return empty.dict

    def set_response_defaults(self, response, request=None):
        """Sets up the response defaults that are defined in the URL route"""
        for header_name, header_value in self.response_headers:
            response.set_header(header_name, header_value)
        if self.set_status:
            response.status = self.set_status
        response.content_type = self.content_type(request, response)

    def render_errors(self, errors, request, response):
        data = {"errors": errors}
        if getattr(self, "on_invalid", False):
            data = self.on_invalid(
                data, **self._arguments(self._params_for_on_invalid, request, response)
            )

        response.status = HTTP_BAD_REQUEST
        if getattr(self, "invalid_outputs", False):
            response.content_type = self.invalid_content_type(request, response)
            response.data = self.invalid_outputs(
                data, **self._arguments(self._params_for_invalid_outputs, request, response)
            )
        else:
            response.data = self.outputs(
                data, **self._arguments(self._params_for_outputs, request, response)
            )

    def call_function(self, parameters):
        if not self.interface.takes_kwargs:
            parameters = {
                key: value for key, value in parameters.items() if key in self.all_parameters
            }
        if getattr(self, "map_params", None):
            self._rewrite_params(parameters)

        return self.interface(**parameters)

    def render_content(self, content, context, request, response, **kwargs):
        if hasattr(content, "interface") and (
            content.interface is True or hasattr(content.interface, "http")
        ):
            if content.interface is True:
                content(request, response, api_version=None, **kwargs)
            else:
                content.interface.http(request, response, api_version=None, **kwargs)
            return

        content = self.transform_data(content, request, response, context)
        content = self.outputs(
            content, **self._arguments(self._params_for_outputs, request, response)
        )
        if hasattr(content, "read"):
            size = None
            if hasattr(content, "name") and os.path.isfile(content.name):
                size = os.path.getsize(content.name)
            if request.range and size:
                start, end = request.range
                if end < 0:
                    end = size + end
                end = min(end, size)
                length = end - start + 1
                content.seek(start)
                response.data = content.read(length)
                response.status = falcon.HTTP_206
                response.content_range = (start, end, size)
                content.close()
            else:
                if size:
                    response.set_stream(content, size)
                else:
                    response.stream = content
        else:
            response.data = content

    def __call__(self, request, response, api_version=None, **kwargs):
        context = self.api.context_factory(
            response=response,
            request=request,
            api=self.api,
            api_version=api_version,
            interface=self,
        )
        """Call the wrapped function over HTTP pulling information as needed"""
        if isinstance(api_version, str) and api_version.isdigit():
            api_version = int(api_version)
        else:
            api_version = None
        if not self.catch_exceptions:
            exception_types = ()
        else:
            exception_types = self.api.http.exception_handlers(api_version)
            exception_types = tuple(exception_types.keys()) if exception_types else ()
        input_parameters = {}
        try:
            self.set_response_defaults(response, request)
            lacks_requirement = self.check_requirements(request, response, context)
            if lacks_requirement:
                response.data = self.outputs(
                    lacks_requirement,
                    **self._arguments(self._params_for_outputs, request, response)
                )
                self.api.delete_context(context, lacks_requirement=lacks_requirement)
                return

            input_parameters = self.gather_parameters(
                request, response, context, api_version, **kwargs
            )
            errors = self.validate(input_parameters, context)
            if errors:
                self.api.delete_context(context, errors=errors)
                return self.render_errors(errors, request, response)

            self.render_content(
                self.call_function(input_parameters), context, request, response, **kwargs
            )
        except falcon.HTTPNotFound as exception:
            self.cleanup_parameters(input_parameters, exception=exception)
            self.api.delete_context(context, exception=exception)
            return self.api.http.not_found(request, response, **kwargs)
        except exception_types as exception:
            self.cleanup_parameters(input_parameters, exception=exception)
            self.api.delete_context(context, exception=exception)
            handler = None
            exception_type = type(exception)
            if exception_type in exception_types:
                handler = self.api.http.exception_handlers(api_version)[exception_type][0]
            else:
                for match_exception_type, exception_handlers in tuple(
                    self.api.http.exception_handlers(api_version).items()
                )[::-1]:
                    if isinstance(exception, match_exception_type):
                        for potential_handler in exception_handlers:
                            if not isinstance(exception, potential_handler.exclude):
                                handler = potential_handler

            if not handler:
                raise exception

            handler(request=request, response=response, exception=exception, **kwargs)
        except Exception as exception:
            self.cleanup_parameters(input_parameters, exception=exception)
            self.api.delete_context(context, exception=exception)
            raise exception
        self.cleanup_parameters(input_parameters)
        self.api.delete_context(context)

    def documentation(self, add_to=None, version=None, prefix="", base_url="", url=""):
        """Returns the documentation specific to an HTTP interface"""
        doc = OrderedDict() if add_to is None else add_to

        usage = self.interface.spec.__doc__
        if usage:
            doc["usage"] = usage

        for example in self.examples:
            example_text = "{0}{1}{2}{3}".format(
                prefix, base_url, "/v{0}".format(version) if version else "", url
            )
            if isinstance(example, str):
                example_text += "?{0}".format(example)
            doc_examples = doc.setdefault("examples", [])
            if not example_text in doc_examples:
                doc_examples.append(example_text)

        doc = super().documentation(doc)

        if getattr(self, "output_doc", ""):
            doc["outputs"]["type"] = self.output_doc

        return doc

    @lru_cache()
    def urls(self, version=None):
        """Returns all URLS that are mapped to this interface"""
        urls = []
        for _base_url, routes in self.api.http.routes.items():
            for url, methods in routes.items():
                for _method, versions in methods.items():
                    for interface_version, interface in versions.items():
                        if interface_version == version and interface == self:
                            if not url in urls:
                                urls.append(("/v{0}".format(version) if version else "") + url)
        return urls

    def url(self, version=None, **kwargs):
        """Returns the first matching URL found for the specified arguments"""
        for url in self.urls(version):
            if [key for key in kwargs.keys() if not "{" + key + "}" in url]:
                continue

            return url.format(**kwargs)

        raise KeyError("URL that takes all provided parameters not found")


class ExceptionRaised(HTTP):
    """Defines the interface responsible for taking and transforming exceptions that occur during processing"""

    __slots__ = ("handle", "exclude")

    def __init__(self, route, *args, **kwargs):
        self.handle = route["exceptions"]
        self.exclude = route["exclude"]
        super().__init__(route, *args, **kwargs)



"""hug/input_formats.py

Defines the built-in Hug input_formatting handlers

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import re
from cgi import parse_multipart
from urllib.parse import parse_qs as urlencoded_converter

from falcon.util.uri import parse_query_string

from hug.format import content_type, underscore
from hug.json_module import json as json_converter


@content_type("text/plain")
def text(body, charset="utf-8", **kwargs):
    """Takes plain text data"""
    return body.read().decode(charset)


@content_type("application/json")
def json(body, charset="utf-8", **kwargs):
    """Takes JSON formatted data, converting it into native Python objects"""
    return json_converter.loads(text(body, charset=charset))


def _underscore_dict(dictionary):
    new_dictionary = {}
    for key, value in dictionary.items():
        if isinstance(value, dict):
            value = _underscore_dict(value)
        if isinstance(key, str):
            key = underscore(key)
        new_dictionary[key] = value
    return new_dictionary


def json_underscore(body, charset="utf-8", **kwargs):
    """Converts JSON formatted date to native Python objects.

    The keys in any JSON dict are transformed from camelcase to underscore separated words.
    """
    return _underscore_dict(json(body, charset=charset))


@content_type("application/x-www-form-urlencoded")
def urlencoded(body, charset="ascii", **kwargs):
    """Converts query strings into native Python objects"""
    return parse_query_string(text(body, charset=charset), False)


@content_type("multipart/form-data")
def multipart(body, content_length=0, **header_params):
    """Converts multipart form data into native Python objects"""
    header_params.setdefault("CONTENT-LENGTH", content_length)
    if header_params and "boundary" in header_params:
        if type(header_params["boundary"]) is str:
            header_params["boundary"] = header_params["boundary"].encode()

    form = parse_multipart((body.stream if hasattr(body, "stream") else body), header_params)
    for key, value in form.items():
        if type(value) is list and len(value) is 1:
            form[key] = value[0]
    return form



"""hug/decorators.py

Defines the method decorators at the core of Hug's approach to creating HTTP APIs

- Decorators for exposing python method as HTTP methods (get, post, etc)
- Decorators for setting the default output and input formats used throughout an API using the framework
- Decorator for registering a new directive method
- Decorator for including another API modules handlers into the current one, with opitonal prefix route

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import functools
from collections import namedtuple

from falcon import HTTP_METHODS

import hug.api
import hug.defaults
import hug.output_format
from hug import introspect
from hug.format import underscore


def default_output_format(
    content_type="application/json", apply_globally=False, api=None, cli=False, http=True
):
    """A decorator that allows you to override the default output format for an API"""

    def decorator(formatter):
        formatter = hug.output_format.content_type(content_type)(formatter)
        if apply_globally:
            if http:
                hug.defaults.output_format = formatter
            if cli:
                hug.defaults.cli_output_format = formatter
        else:
            apply_to_api = hug.API(api) if api else hug.api.from_object(formatter)
            if http:
                apply_to_api.http.output_format = formatter
            if cli:
                apply_to_api.cli.output_format = formatter
        return formatter

    return decorator


def default_input_format(content_type="application/json", apply_globally=False, api=None):
    """A decorator that allows you to override the default output format for an API"""

    def decorator(formatter):
        formatter = hug.output_format.content_type(content_type)(formatter)
        if apply_globally:
            hug.defaults.input_format[content_type] = formatter
        else:
            apply_to_api = hug.API(api) if api else hug.api.from_object(formatter)
            apply_to_api.http.set_input_format(content_type, formatter)
        return formatter

    return decorator


def directive(apply_globally=False, api=None):
    """A decorator that registers a single hug directive"""

    def decorator(directive_method):
        if apply_globally:
            hug.defaults.directives[underscore(directive_method.__name__)] = directive_method
        else:
            apply_to_api = hug.API(api) if api else hug.api.from_object(directive_method)
            apply_to_api.add_directive(directive_method)
        directive_method.directive = True
        return directive_method

    return decorator


def context_factory(apply_globally=False, api=None):
    """A decorator that registers a single hug context factory"""

    def decorator(context_factory_):
        if apply_globally:
            hug.defaults.context_factory = context_factory_
        else:
            apply_to_api = hug.API(api) if api else hug.api.from_object(context_factory_)
            apply_to_api.context_factory = context_factory_
        return context_factory_

    return decorator


def delete_context(apply_globally=False, api=None):
    """A decorator that registers a single hug delete context function"""

    def decorator(delete_context_):
        if apply_globally:
            hug.defaults.delete_context = delete_context_
        else:
            apply_to_api = hug.API(api) if api else hug.api.from_object(delete_context_)
            apply_to_api.delete_context = delete_context_
        return delete_context_

    return decorator


def startup(api=None):
    """Runs the provided function on startup, passing in an instance of the api"""

    def startup_wrapper(startup_function):
        apply_to_api = hug.API(api) if api else hug.api.from_object(startup_function)
        apply_to_api.add_startup_handler(startup_function)
        return startup_function

    return startup_wrapper


def request_middleware(api=None):
    """Registers a middleware function that will be called on every request"""

    def decorator(middleware_method):
        apply_to_api = hug.API(api) if api else hug.api.from_object(middleware_method)

        class MiddlewareRouter(object):
            __slots__ = ()

            def process_request(self, request, response):
                return middleware_method(request, response)

        apply_to_api.http.add_middleware(MiddlewareRouter())
        return middleware_method

    return decorator


def response_middleware(api=None):
    """Registers a middleware function that will be called on every response"""

    def decorator(middleware_method):
        apply_to_api = hug.API(api) if api else hug.api.from_object(middleware_method)

        class MiddlewareRouter(object):
            __slots__ = ()

            def process_response(self, request, response, resource, _req_succeeded):
                return middleware_method(request, response, resource)

        apply_to_api.http.add_middleware(MiddlewareRouter())
        return middleware_method

    return decorator


def reqresp_middleware(api=None):
    """Registers a middleware function that will be called on every request and response"""

    def decorator(middleware_generator):
        apply_to_api = hug.API(api) if api else hug.api.from_object(middleware_generator)

        class MiddlewareRouter(object):
            __slots__ = ("gen",)

            def process_request(self, request, response):
                self.gen = middleware_generator(request)
                return self.gen.__next__()

            def process_response(self, request, response, resource, _req_succeeded):
                return self.gen.send((response, resource))

        apply_to_api.http.add_middleware(MiddlewareRouter())
        return middleware_generator

    return decorator


def middleware_class(api=None):
    """Registers a middleware class"""

    def decorator(middleware_class):
        apply_to_api = hug.API(api) if api else hug.api.from_object(middleware_class)
        apply_to_api.http.add_middleware(middleware_class())
        return middleware_class

    return decorator


def extend_api(route="", api=None, base_url="", **kwargs):
    """Extends the current api, with handlers from an imported api. Optionally provide a route that prefixes access"""

    def decorator(extend_with):
        apply_to_api = hug.API(api) if api else hug.api.from_object(extend_with)
        for extended_api in extend_with():
            apply_to_api.extend(extended_api, route, base_url, **kwargs)
        return extend_with

    return decorator


def wraps(function):
    """Enables building decorators around functions used for hug routes without changing their function signature"""

    def wrap(decorator):
        decorator = functools.wraps(function)(decorator)
        if not hasattr(function, "original"):
            decorator.original = function
        else:
            decorator.original = function.original
            delattr(function, "original")
        return decorator

    return wrap


def auto_kwargs(function):
    """Modifies the provided function to support kwargs by only passing along kwargs for parameters it accepts"""
    supported = introspect.arguments(function)

    @wraps(function)
    def call_function(*args, **kwargs):
        return function(*args, **{key: value for key, value in kwargs.items() if key in supported})

    return call_function



"""hug/directives.py

Defines the directives built into hug. Directives allow attaching behaviour to an API handler based simply
on an argument it takes and that arguments default value. The directive gets called with the default supplied,
ther request data, and api_version. The result of running the directive method is then set as the argument value.
Directive attributes are always prefixed with 'hug_'

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

from functools import partial
from timeit import default_timer as python_timer

from hug import introspect


def _built_in_directive(directive):
    """Marks a callable as a built-in directive"""
    directive.directive = True
    return directive


@_built_in_directive
class Timer(object):
    """Keeps track of time surpased since instantiation, outputed by doing float(instance)"""

    __slots__ = ("start", "round_to")

    def __init__(self, round_to=None, **kwargs):
        self.start = python_timer()
        self.round_to = round_to

    def __float__(self):
        time_taken = python_timer() - self.start
        return round(time_taken, self.round_to) if self.round_to else time_taken

    def __int__(self):
        return int(round(float(self)))

    def __native_types__(self):
        return self.__float__()

    def __str__(self):
        return str(float(self))

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self)


@_built_in_directive
def module(default=None, api=None, **kwargs):
    """Returns the module that is running this hug API function"""
    return api.module if api else default


@_built_in_directive
def api(default=None, api=None, **kwargs):
    """Returns the api instance in which this API function is being ran"""
    return api if api else default


@_built_in_directive
def api_version(default=None, api_version=None, **kwargs):
    """Returns the current api_version as a directive for use in both request and not request handling code"""
    return api_version


@_built_in_directive
def documentation(default=None, api_version=None, api=None, **kwargs):
    """returns documentation for the current api"""
    api_version = default or api_version
    if api:
        return api.http.documentation(base_url="", api_version=api_version)


@_built_in_directive
def session(context_name="session", request=None, **kwargs):
    """Returns the session associated with the current request"""
    return request and request.context.get(context_name, None)


@_built_in_directive
def user(default=None, request=None, **kwargs):
    """Returns the current logged in user"""
    return request and request.context.get("user", None) or default


@_built_in_directive
def cors(support="*", response=None, **kwargs):
    """Adds the the Access-Control-Allow-Origin header to this endpoint, with the specified support"""
    response and response.set_header("Access-Control-Allow-Origin", support)
    return support


@_built_in_directive
class CurrentAPI(object):
    """Returns quick access to all api functions on the current version of the api"""

    __slots__ = ("api_version", "api")

    def __init__(self, default=None, api_version=None, **kwargs):
        self.api_version = api_version
        self.api = api(**kwargs)

    def __getattr__(self, name):
        function = self.api.http.versioned.get(self.api_version, {}).get(name, None)
        if not function:
            function = self.api.http.versioned.get(None, {}).get(name, None)
        if not function:
            raise AttributeError("API Function {0} not found".format(name))

        accepts = function.interface.arguments
        if "hug_api_version" in accepts:
            function = partial(function, hug_api_version=self.api_version)
        if "hug_current_api" in accepts:
            function = partial(function, hug_current_api=self)

        return function



import os

HUG_USE_UJSON = bool(os.environ.get("HUG_USE_UJSON", 1))
try:  # pragma: no cover
    if HUG_USE_UJSON:
        import ujson as json

        class dumps_proxy:  # noqa: N801
            """Proxies the call so non supported kwargs are skipped
            and it enables escape_forward_slashes to simulate built-in json
            """

            _dumps = json.dumps

            def __call__(self, *args, **kwargs):
                kwargs.pop("default", None)
                kwargs.pop("separators", None)
                kwargs.update(escape_forward_slashes=False)
                try:
                    return self._dumps(*args, **kwargs)
                except Exception as exc:
                    raise TypeError("Type[ujson] is not Serializable", exc)

        json.dumps = dumps_proxy()
    else:
        import json
except ImportError:  # pragma: no cover
    import json



"""hug/introspect.py

Defines built in hug functions to aid in introspection

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import inspect
from types import MethodType


def is_method(function):
    """Returns True if the passed in function is identified as a method (NOT a function)"""
    return isinstance(function, MethodType)


def is_coroutine(function):
    """Returns True if the passed in function is a coroutine"""
    return function.__code__.co_flags & 0x0080 or getattr(function, "_is_coroutine", False)


def name(function):
    """Returns the name of a function"""
    return function.__name__


def arguments(function, extra_arguments=0):
    """Returns the name of all arguments a function takes"""
    if not hasattr(function, "__code__"):
        return ()

    return function.__code__.co_varnames[: function.__code__.co_argcount + extra_arguments]


def takes_kwargs(function):
    """Returns True if the supplied function takes keyword arguments"""
    return bool(function.__code__.co_flags & 0x08)


def takes_args(function):
    """Returns True if the supplied functions takes extra non-keyword arguments"""
    return bool(function.__code__.co_flags & 0x04)


def takes_arguments(function, *named_arguments):
    """Returns the arguments that a function takes from a list of requested arguments"""
    return set(named_arguments).intersection(arguments(function))


def takes_all_arguments(function, *named_arguments):
    """Returns True if all supplied arguments are found in the function"""
    return bool(takes_arguments(function, *named_arguments) == set(named_arguments))


def generate_accepted_kwargs(function, *named_arguments):
    """Dynamically creates a function that when called with dictionary of arguments will produce a kwarg that's
       compatible with the supplied function
    """
    if hasattr(function, "__code__") and takes_kwargs(function):
        function_takes_kwargs = True
        function_takes_arguments = []
    else:
        function_takes_kwargs = False
        function_takes_arguments = takes_arguments(function, *named_arguments)

    def accepted_kwargs(kwargs):
        if function_takes_kwargs:
            return kwargs
        elif function_takes_arguments:
            return {key: value for key, value in kwargs.items() if key in function_takes_arguments}
        return {}

    return accepted_kwargs



"""hug/store.py.

A collecton of native stores which can be used with, among others, the session middleware.

Copyright (C) 2016 Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from hug.exceptions import StoreKeyNotFound


class InMemoryStore:
    """
    Naive store class which can be used for the session middleware and unit tests.
    It is not thread-safe and no data will survive the lifecycle of the hug process.
    Regard this as a blueprint for more useful and probably more complex store implementations, for example stores
    which make use of databases like Redis, PostgreSQL or others.
    """

    def __init__(self):
        self._data = {}

    def get(self, key):
        """Get data for given store key. Raise hug.exceptions.StoreKeyNotFound if key does not exist."""
        try:
            data = self._data[key]
        except KeyError:
            raise StoreKeyNotFound(key)
        return data

    def exists(self, key):
        """Return whether key exists or not."""
        return key in self._data

    def set(self, key, data):
        """Set data object for given store key."""
        self._data[key] = data

    def delete(self, key):
        """Delete data for given store key."""
        if key in self._data:
            del self._data[key]



"""hug/types.py

Defines hugs built-in supported types / validators

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import uuid as native_uuid
from decimal import Decimal
from distutils.version import LooseVersion

import hug._empty as empty
from hug import introspect
from hug.exceptions import InvalidTypeData
from hug.json_module import json as json_converter

MARSHMALLOW_MAJOR_VERSION = None
try:
    import marshmallow
    from marshmallow import ValidationError

    MARSHMALLOW_MAJOR_VERSION = getattr(
        marshmallow, "__version_info__", LooseVersion(marshmallow.__version__).version
    )[0]
except ImportError:
    # Just define the error that is never raised so that Python does not complain.
    class ValidationError(Exception):
        pass


class Type(object):
    """Defines the base hug concept of a type for use in function annotation.
       Override `__call__` to define how the type should be transformed and validated
    """

    _hug_type = True
    _sub_type = None
    _accept_context = False

    def __init__(self):
        pass

    def __call__(self, value):
        raise NotImplementedError("To implement a new type __call__ must be defined")


def create(
    doc=None,
    error_text=None,
    exception_handlers=empty.dict,
    extend=Type,
    chain=True,
    auto_instance=True,
    accept_context=False,
):
    """Creates a new type handler with the specified type-casting handler"""
    extend = extend if type(extend) == type else type(extend)

    def new_type_handler(function):
        class NewType(extend):
            __slots__ = ()
            _accept_context = accept_context

            if chain and extend != Type:
                if error_text or exception_handlers:
                    if not accept_context:

                        def __call__(self, value):
                            try:
                                value = super(NewType, self).__call__(value)
                                return function(value)
                            except Exception as exception:
                                for take_exception, rewrite in exception_handlers.items():
                                    if isinstance(exception, take_exception):
                                        if isinstance(rewrite, str):
                                            raise ValueError(rewrite)
                                        else:
                                            raise rewrite(value)
                                if error_text:
                                    raise ValueError(error_text)
                                raise exception

                    else:
                        if extend._accept_context:

                            def __call__(self, value, context):
                                try:
                                    value = super(NewType, self).__call__(value, context)
                                    return function(value, context)
                                except Exception as exception:
                                    for take_exception, rewrite in exception_handlers.items():
                                        if isinstance(exception, take_exception):
                                            if isinstance(rewrite, str):
                                                raise ValueError(rewrite)
                                            else:
                                                raise rewrite(value)
                                    if error_text:
                                        raise ValueError(error_text)
                                    raise exception

                        else:

                            def __call__(self, value, context):
                                try:
                                    value = super(NewType, self).__call__(value)
                                    return function(value, context)
                                except Exception as exception:
                                    for take_exception, rewrite in exception_handlers.items():
                                        if isinstance(exception, take_exception):
                                            if isinstance(rewrite, str):
                                                raise ValueError(rewrite)
                                            else:
                                                raise rewrite(value)
                                    if error_text:
                                        raise ValueError(error_text)
                                    raise exception

                else:
                    if not accept_context:

                        def __call__(self, value):
                            value = super(NewType, self).__call__(value)
                            return function(value)

                    else:
                        if extend._accept_context:

                            def __call__(self, value, context):
                                value = super(NewType, self).__call__(value, context)
                                return function(value, context)

                        else:

                            def __call__(self, value, context):
                                value = super(NewType, self).__call__(value)
                                return function(value, context)

            else:
                if not accept_context:
                    if error_text or exception_handlers:

                        def __call__(self, value):
                            try:
                                return function(value)
                            except Exception as exception:
                                for take_exception, rewrite in exception_handlers.items():
                                    if isinstance(exception, take_exception):
                                        if isinstance(rewrite, str):
                                            raise ValueError(rewrite)
                                        else:
                                            raise rewrite(value)
                                if error_text:
                                    raise ValueError(error_text)
                                raise exception

                    else:

                        def __call__(self, value):
                            return function(value)

                else:
                    if error_text or exception_handlers:

                        def __call__(self, value, context):
                            try:
                                return function(value, context)
                            except Exception as exception:
                                for take_exception, rewrite in exception_handlers.items():
                                    if isinstance(exception, take_exception):
                                        if isinstance(rewrite, str):
                                            raise ValueError(rewrite)
                                        else:
                                            raise rewrite(value)
                                if error_text:
                                    raise ValueError(error_text)
                                raise exception

                    else:

                        def __call__(self, value, context):
                            return function(value, context)

        NewType.__doc__ = function.__doc__ if doc is None else doc
        if auto_instance and not (
            introspect.arguments(NewType.__init__, -1)
            or introspect.takes_kwargs(NewType.__init__)
            or introspect.takes_args(NewType.__init__)
        ):
            return NewType()
        return NewType

    return new_type_handler


def accept(kind, doc=None, error_text=None, exception_handlers=empty.dict, accept_context=False):
    """Allows quick wrapping of any Python type cast function for use as a hug type annotation"""
    return create(
        doc,
        error_text,
        exception_handlers=exception_handlers,
        chain=False,
        accept_context=accept_context,
    )(kind)


number = accept(int, "A whole number", "Invalid whole number provided")
float_number = accept(float, "A float number", "Invalid float number provided")
decimal = accept(Decimal, "A decimal number", "Invalid decimal number provided")
boolean = accept(
    bool, "Providing any value will set this to true", "Invalid boolean value provided"
)
uuid = accept(native_uuid.UUID, "A Universally Unique IDentifier", "Invalid UUID provided")


class Text(Type):
    """Basic text / string value"""

    __slots__ = ()

    def __call__(self, value):
        if type(value) in (list, tuple) or value is None:
            raise ValueError("Invalid text value provided")
        return str(value)


text = Text()


class SubTyped(type):
    def __getitem__(cls, sub_type):
        class TypedSubclass(cls):
            _sub_type = sub_type

        return TypedSubclass


class Multiple(Type, metaclass=SubTyped):
    """Multiple Values"""

    __slots__ = ()

    def __call__(self, value):
        as_multiple = value if isinstance(value, list) else [value]
        if self._sub_type:
            return [self._sub_type(item) for item in as_multiple]
        return as_multiple


class DelimitedList(Type, metaclass=SubTyped):
    """Defines a list type that is formed by delimiting a list with a certain character or set of characters"""

    def __init__(self, using=","):
        super().__init__()
        self.using = using

    @property
    def __doc__(self):
        return '''Multiple values, separated by "{0}"'''.format(self.using)

    def __call__(self, value):
        value_list = value if type(value) in (list, tuple) else value.split(self.using)
        if self._sub_type:
            value_list = [self._sub_type(val) for val in value_list]
        return value_list


class SmartBoolean(type(boolean)):
    """Accepts a true or false value"""

    __slots__ = ()

    def __call__(self, value):
        if type(value) == bool or value in (None, 1, 0):
            return bool(value)

        value = value.lower()
        if value in ("true", "t", "1"):
            return True
        elif value in ("false", "f", "0", ""):
            return False

        raise KeyError("Invalid value passed in for true/false field")


class InlineDictionary(Type, metaclass=SubTyped):
    """A single line dictionary, where items are separted by commas and key:value are separated by a pipe"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key_type = self.value_type = None
        if self._sub_type:
            if type(self._sub_type) in (tuple, list):
                if len(self._sub_type) >= 2:
                    self.key_type, self.value_type = self._sub_type[:2]
            else:
                self.key_type = self._sub_type

    def __call__(self, string):
        dictionary = {}
        for key, value in (item.split(":") for item in string.split("|")):
            key, value = key.strip(), value.strip()
            dictionary[self.key_type(key) if self.key_type else key] = (
                self.value_type(value) if self.value_type else value
            )
        return dictionary


class OneOf(Type):
    """Ensures the value is within a set of acceptable values"""

    __slots__ = ("values",)

    def __init__(self, values):
        self.values = values

    @property
    def __doc__(self):
        return "Accepts one of the following values: ({0})".format("|".join(self.values))

    def __call__(self, value):
        if not value in self.values:
            raise KeyError(
                "Invalid value passed. The accepted values are: ({0})".format("|".join(self.values))
            )
        return value


class Mapping(OneOf):
    """Ensures the value is one of an acceptable set of values mapping those values to a Python equivelent"""

    __slots__ = ("value_map",)

    def __init__(self, value_map):
        self.value_map = value_map
        self.values = value_map.keys()

    @property
    def __doc__(self):
        return "Accepts one of the following values: ({0})".format("|".join(self.values))

    def __call__(self, value):
        if not value in self.values:
            raise KeyError(
                "Invalid value passed. The accepted values are: ({0})".format("|".join(self.values))
            )
        return self.value_map[value]


class JSON(Type):
    """Accepts a JSON formatted data structure"""

    __slots__ = ()

    def __call__(self, value):
        if type(value) in (str, bytes):
            try:
                return json_converter.loads(value)
            except Exception:
                raise ValueError("Incorrectly formatted JSON provided")
        if type(value) is list:
            # If Falcon is set to comma-separate entries, this segment joins them again.
            try:
                fixed_value = ",".join(value)
                return json_converter.loads(fixed_value)
            except Exception:
                raise ValueError("Incorrectly formatted JSON provided")
        else:
            return value


class Multi(Type):
    """Enables accepting one of multiple type methods"""

    __slots__ = ("types",)

    def __init__(self, *types):
        self.types = types

    @property
    def __doc__(self):
        type_strings = (type_method.__doc__ for type_method in self.types)
        return "Accepts any of the following value types:{0}\n".format("\n  - ".join(type_strings))

    def __call__(self, value):
        for type_method in self.types:
            try:
                return type_method(value)
            except BaseException:
                pass
        raise ValueError(self.__doc__)


class InRange(Type):
    """Accepts a number within a lower and upper bound of acceptable values"""

    __slots__ = ("lower", "upper", "convert")

    def __init__(self, lower, upper, convert=number):
        self.lower = lower
        self.upper = upper
        self.convert = convert

    @property
    def __doc__(self):
        return "{0} that is greater or equal to {1} and less than {2}".format(
            self.convert.__doc__, self.lower, self.upper
        )

    def __call__(self, value):
        value = self.convert(value)
        if value < self.lower:
            raise ValueError("'{0}' is less than the lower limit {1}".format(value, self.lower))
        if value >= self.upper:
            raise ValueError("'{0}' reaches the limit of {1}".format(value, self.upper))
        return value


class LessThan(Type):
    """Accepts a number within a lower and upper bound of acceptable values"""

    __slots__ = ("limit", "convert")

    def __init__(self, limit, convert=number):
        self.limit = limit
        self.convert = convert

    @property
    def __doc__(self):
        return "{0} that is less than {1}".format(self.convert.__doc__, self.limit)

    def __call__(self, value):
        value = self.convert(value)
        if not value < self.limit:
            raise ValueError("'{0}' must be less than {1}".format(value, self.limit))
        return value


class GreaterThan(Type):
    """Accepts a value above a given minimum"""

    __slots__ = ("minimum", "convert")

    def __init__(self, minimum, convert=number):
        self.minimum = minimum
        self.convert = convert

    @property
    def __doc__(self):
        return "{0} that is greater than {1}".format(self.convert.__doc__, self.minimum)

    def __call__(self, value):
        value = self.convert(value)
        if not value > self.minimum:
            raise ValueError("'{0}' must be greater than {1}".format(value, self.minimum))
        return value


class Length(Type):
    """Accepts a a value that is within a specific length limit"""

    __slots__ = ("lower", "upper", "convert")

    def __init__(self, lower, upper, convert=text):
        self.lower = lower
        self.upper = upper
        self.convert = convert

    @property
    def __doc__(self):
        return "{0} that has a length longer or equal to {1} and less then {2}".format(
            self.convert.__doc__, self.lower, self.upper
        )

    def __call__(self, value):
        value = self.convert(value)
        length = len(value)
        if length < self.lower:
            raise ValueError(
                "'{0}' is shorter than the lower limit of {1}".format(value, self.lower)
            )
        if length >= self.upper:
            raise ValueError(
                "'{0}' is longer then the allowed limit of {1}".format(value, self.upper)
            )
        return value


class ShorterThan(Type):
    """Accepts a text value shorter than the specified length limit"""

    __slots__ = ("limit", "convert")

    def __init__(self, limit, convert=text):
        self.limit = limit
        self.convert = convert

    @property
    def __doc__(self):
        return "{0} with a length of no more than {1}".format(self.convert.__doc__, self.limit)

    def __call__(self, value):
        value = self.convert(value)
        length = len(value)
        if not length < self.limit:
            raise ValueError(
                "'{0}' is longer then the allowed limit of {1}".format(value, self.limit)
            )
        return value


class LongerThan(Type):
    """Accepts a value up to the specified limit"""

    __slots__ = ("limit", "convert")

    def __init__(self, limit, convert=text):
        self.limit = limit
        self.convert = convert

    @property
    def __doc__(self):
        return "{0} with a length longer than {1}".format(self.convert.__doc__, self.limit)

    def __call__(self, value):
        value = self.convert(value)
        length = len(value)
        if not length > self.limit:
            raise ValueError("'{0}' must be longer than {1}".format(value, self.limit))
        return value


class CutOff(Type):
    """Cuts off the provided value at the specified index"""

    __slots__ = ("limit", "convert")

    def __init__(self, limit, convert=text):
        self.limit = limit
        self.convert = convert

    @property
    def __doc__(self):
        return "'{0}' with anything over the length of {1} being ignored".format(
            self.convert.__doc__, self.limit
        )

    def __call__(self, value):
        return self.convert(value)[: self.limit]


class Chain(Type):
    """type for chaining multiple types together"""

    __slots__ = ("types",)

    def __init__(self, *types):
        self.types = types

    def __call__(self, value):
        for type_function in self.types:
            value = type_function(value)
        return value


class Nullable(Chain):
    """A Chain types that Allows None values"""

    __slots__ = ("types",)

    def __init__(self, *types):
        self.types = types

    def __call__(self, value):
        if value is None:
            return None
        else:
            return super(Nullable, self).__call__(value)


class TypedProperty(object):
    """class for building property objects for schema objects"""

    __slots__ = ("name", "type_func")

    def __init__(self, name, type_func):
        self.name = "_" + name
        self.type_func = type_func

    def __get__(self, instance, cls):
        return getattr(instance, self.name, None)

    def __set__(self, instance, value):
        setattr(instance, self.name, self.type_func(value))

    def __delete__(self, instance):
        raise AttributeError("Can't delete attribute")


class NewTypeMeta(type):
    """Meta class to turn Schema objects into format usable by hug"""

    __slots__ = ()

    def __init__(cls, name, bases, nmspc):
        cls._types = {
            attr: getattr(cls, attr)
            for attr in dir(cls)
            if getattr(getattr(cls, attr), "_hug_type", False)
        }
        slots = getattr(cls, "__slots__", ())
        slots = set(slots)
        for attr, type_func in cls._types.items():
            slots.add("_" + attr)
            slots.add(attr)
            prop = TypedProperty(attr, type_func)
            setattr(cls, attr, prop)
        cls.__slots__ = tuple(slots)
        super(NewTypeMeta, cls).__init__(name, bases, nmspc)


class Schema(object, metaclass=NewTypeMeta):
    """Schema for creating complex types using hug types"""

    __slots__ = ()

    def __new__(cls, json, *args, **kwargs):
        if json.__class__ == cls:
            return json
        else:
            return super(Schema, cls).__new__(cls)

    def __init__(self, json, force=False):
        if self != json:
            for (key, value) in json.items():
                if force:
                    key = "_" + key
                setattr(self, key, value)


json = JSON()


class MarshmallowInputSchema(Type):
    """Allows using a Marshmallow Schema directly in a hug input type annotation"""

    __slots__ = "schema"

    def __init__(self, schema):
        self.schema = schema

    @property
    def __doc__(self):
        return self.schema.__doc__ or self.schema.__class__.__name__

    def __call__(self, value, context):
        self.schema.context = context
        # In marshmallow 2 schemas return tuple (`data`, `errors`) upon loading. They might also raise on invalid data
        # if configured so, but will still return a tuple.
        # In marshmallow 3 schemas always raise Validation error on load if input data is invalid and a single
        # value `data` is returned.
        if MARSHMALLOW_MAJOR_VERSION is None or MARSHMALLOW_MAJOR_VERSION == 2:
            value, errors = (
                self.schema.loads(value) if isinstance(value, str) else self.schema.load(value)
            )
        else:
            errors = {}
            try:
                value = (
                    self.schema.loads(value) if isinstance(value, str) else self.schema.load(value)
                )
            except ValidationError as e:
                errors = e.messages

        if errors:
            raise InvalidTypeData(
                "Invalid {0} passed in".format(self.schema.__class__.__name__), errors
            )
        return value


class MarshmallowReturnSchema(Type):
    """Allows using a Marshmallow Schema directly in a hug return type annotation"""

    __slots__ = ("schema",)

    def __init__(self, schema):
        self.schema = schema

    @property
    def context(self):
        return self.schema.context

    @context.setter
    def context(self, context):
        self.schema.context = context

    @property
    def __doc__(self):
        return self.schema.__doc__ or self.schema.__class__.__name__

    def __call__(self, value):
        # In marshmallow 2 schemas return tuple (`data`, `errors`) upon loading. They might also raise on invalid data
        # if configured so, but will still return a tuple.
        # In marshmallow 3 schemas always raise Validation error on load if input data is invalid and a single
        # value `data` is returned.
        if MARSHMALLOW_MAJOR_VERSION is None or MARSHMALLOW_MAJOR_VERSION == 2:
            value, errors = self.schema.dump(value)
        else:
            errors = {}
            try:
                value = self.schema.dump(value)
            except ValidationError as e:
                errors = e.messages

        if errors:
            raise InvalidTypeData(
                "Invalid {0} passed in".format(self.schema.__class__.__name__), errors
            )
        return value


multiple = Multiple()
smart_boolean = SmartBoolean()
inline_dictionary = InlineDictionary()
comma_separated_list = DelimitedList(using=",")


# NOTE: These forms are going to be DEPRECATED, here for backwards compatibility only
delimited_list = DelimitedList
one_of = OneOf
mapping = Mapping
multi = Multi
in_range = InRange
less_than = LessThan
greater_than = GreaterThan
length = Length
shorter_than = ShorterThan
longer_than = LongerThan
cut_off = CutOff



"""hug/validate.py

Defines hugs built-in validation methods

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import


def all(*validators):
    """Validation only succeeds if all passed in validators return no errors"""

    def validate_all(fields):
        for validator in validators:
            errors = validator(fields)
            if errors:
                return errors

    validate_all.__doc__ = " and ".join(validator.__doc__ for validator in validators)
    return validate_all


def any(*validators):
    """If any of the specified validators pass the validation succeeds"""

    def validate_any(fields):
        errors = {}
        for validator in validators:
            validation_errors = validator(fields)
            if not validation_errors:
                return
            errors.update(validation_errors)
        return errors

    validate_any.__doc__ = " or ".join(validator.__doc__ for validator in validators)
    return validate_any


def contains_one_of(*fields):
    """Enables ensuring that one of multiple optional fields is set"""
    message = "Must contain any one of the following fields: {0}".format(", ".join(fields))

    def check_contains(endpoint_fields):
        for field in fields:
            if field in endpoint_fields:
                return

        errors = {}
        for field in fields:
            errors[field] = "one of these must have a value"
        return errors

    check_contains.__doc__ = message
    return check_contains



"""hug/authentication.py

Provides the basic built-in authentication helper functions

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

import base64
import binascii

from falcon import HTTPUnauthorized


def authenticator(function, challenges=()):
    """Wraps authentication logic, verify_user through to the authentication function.

    The verify_user function passed in should accept an API key and return a user object to
    store in the request context if authentication succeeded.
    """
    challenges = challenges or ('{} realm="simple"'.format(function.__name__),)

    def wrapper(verify_user):
        def authenticate(request, response, **kwargs):
            result = function(request, response, verify_user, **kwargs)

            def authenticator_name():
                try:
                    return function.__doc__.splitlines()[0]
                except AttributeError:
                    return function.__name__

            if result is None:
                raise HTTPUnauthorized(
                    "Authentication Required",
                    "Please provide valid {0} credentials".format(authenticator_name()),
                    challenges=challenges,
                )

            if result is False:
                raise HTTPUnauthorized(
                    "Invalid Authentication",
                    "Provided {0} credentials were invalid".format(authenticator_name()),
                    challenges=challenges,
                )

            request.context["user"] = result
            return True

        authenticate.__doc__ = function.__doc__
        return authenticate

    return wrapper


@authenticator
def basic(request, response, verify_user, realm="simple", context=None, **kwargs):
    """Basic HTTP Authentication"""
    http_auth = request.auth
    response.set_header("WWW-Authenticate", "Basic")
    if http_auth is None:
        return

    if isinstance(http_auth, bytes):
        http_auth = http_auth.decode("utf8")
    try:
        auth_type, user_and_key = http_auth.split(" ", 1)
    except ValueError:
        raise HTTPUnauthorized(
            "Authentication Error",
            "Authentication header is improperly formed",
            challenges=('Basic realm="{}"'.format(realm),),
        )

    if auth_type.lower() == "basic":
        try:
            user_id, key = (
                base64.decodebytes(bytes(user_and_key.strip(), "utf8")).decode("utf8").split(":", 1)
            )
            try:
                user = verify_user(user_id, key)
            except TypeError:
                user = verify_user(user_id, key, context)
            if user:
                response.set_header("WWW-Authenticate", "")
                return user
        except (binascii.Error, ValueError):
            raise HTTPUnauthorized(
                "Authentication Error",
                "Unable to determine user and password with provided encoding",
                challenges=('Basic realm="{}"'.format(realm),),
            )
    return False


@authenticator
def api_key(request, response, verify_user, context=None, **kwargs):
    """API Key Header Authentication

    The verify_user function passed in to ths authenticator shall receive an
    API key as input, and return a user object to store in the request context
    if the request was successful.
    """
    api_key = request.get_header("X-Api-Key")

    if api_key:
        try:
            user = verify_user(api_key)
        except TypeError:
            user = verify_user(api_key, context)
        if user:
            return user
        else:
            return False
    else:
        return None


@authenticator
def token(request, response, verify_user, context=None, **kwargs):
    """Token verification

    Checks for the Authorization header and verifies using the verify_user function
    """
    token = request.get_header("Authorization")
    if token:
        try:
            verified_token = verify_user(token)
        except TypeError:
            verified_token = verify_user(token, context)
        if verified_token:
            return verified_token
        else:
            return False
    return None


def verify(user, password):
    """Returns a simple verification callback that simply verifies that the users and password match that provided"""

    def verify_user(user_name, user_password):
        if user_name == user and user_password == password:
            return user_name
        return False

    return verify_user



"""hug/development_runner.py

Contains logic to enable execution of hug APIS locally from the command line for development use

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""
from __future__ import absolute_import

import importlib
import os
import subprocess
import sys
import tempfile
import time
from multiprocessing import Process
from os.path import exists

import _thread as thread
from hug._version import current
from hug.api import API
from hug.route import cli
from hug.types import boolean, number

INIT_MODULES = list(sys.modules.keys())


def _start_api(api_module, host, port, no_404_documentation, show_intro=True):
    API(api_module).http.serve(host, port, no_404_documentation, show_intro)


@cli(version=current)
def hug(
    file: "A Python file that contains a Hug API" = None,
    module: "A Python module that contains a Hug API" = None,
    host: "Interface to bind to" = "",
    port: number = 8000,
    no_404_documentation: boolean = False,
    manual_reload: boolean = False,
    interval: number = 1,
    command: "Run a command defined in the given module" = None,
    silent: boolean = False,
):
    """Hug API Development Server"""
    api_module = None
    if file and module:
        print("Error: can not define both a file and module source for Hug API.")
        sys.exit(1)
    if file:
        sys.path.append(os.path.dirname(os.path.abspath(file)))
        sys.path.append(os.getcwd())
        api_module = importlib.machinery.SourceFileLoader(file.split(".")[0], file).load_module()
    elif module:
        sys.path.append(os.getcwd())
        api_module = importlib.import_module(module)
    if not api_module or not hasattr(api_module, "__hug__"):
        print("Error: must define a file name or module that contains a Hug API.")
        sys.exit(1)

    api = API(api_module, display_intro=not silent)
    if command:
        if command not in api.cli.commands:
            print(str(api.cli))
            sys.exit(1)

        flag_index = (sys.argv.index("-c") if "-c" in sys.argv else sys.argv.index("--command")) + 1
        sys.argv = sys.argv[flag_index:]
        api.cli.commands[command]()
        return

    ran = False
    if not manual_reload:
        thread.start_new_thread(reload_checker, (interval,))
        while True:
            reload_checker.reloading = False
            time.sleep(1)
            try:
                _start_api(
                    api_module, host, port, no_404_documentation, False if silent else not ran
                )
            except KeyboardInterrupt:
                if not reload_checker.reloading:
                    sys.exit(1)
                reload_checker.reloading = False
                ran = True
                for name in list(sys.modules.keys()):
                    if name not in INIT_MODULES:
                        del sys.modules[name]
                if file:
                    api_module = importlib.machinery.SourceFileLoader(
                        file.split(".")[0], file
                    ).load_module()
                elif module:
                    api_module = importlib.import_module(module)
    else:
        _start_api(api_module, host, port, no_404_documentation, not ran)


def reload_checker(interval):
    while True:
        changed = False
        files = {}
        for module in list(sys.modules.values()):
            path = getattr(module, "__file__", "")
            if not path:
                continue
            if path[-4:] in (".pyo", ".pyc"):
                path = path[:-1]
            if path and exists(path):
                files[path] = os.stat(path).st_mtime

        while not changed:
            for path, last_modified in files.items():
                if not exists(path):
                    print("\n> Reloading due to file removal: {}".format(path))
                    changed = True
                elif os.stat(path).st_mtime > last_modified:
                    print("\n> Reloading due to file change: {}".format(path))
                    changed = True

                if changed:
                    reload_checker.reloading = True
                    thread.interrupt_main()
                    time.sleep(5)
                    break
            time.sleep(interval)



"""hug/transform.py

Defines Hug's built-in output transforming functions

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

from hug.decorators import auto_kwargs


def content_type(transformers, default=None):
    """Returns a different transformer depending on the content type passed in.
       If none match and no default is given no transformation takes place.

       should pass in a dict with the following format:

            {'[content-type]': transformation_action,
             ...
            }
    """
    transformers = {
        content_type: auto_kwargs(transformer) if transformer else transformer
        for content_type, transformer in transformers.items()
    }
    default = default and auto_kwargs(default)

    def transform(data, request):
        transformer = transformers.get(request.content_type.split(";")[0], default)
        if not transformer:
            return data

        return transformer(data)

    return transform


def suffix(transformers, default=None):
    """Returns a different transformer depending on the suffix at the end of the requested URL.
       If none match and no default is given no transformation takes place.

       should pass in a dict with the following format:

            {'[suffix]': transformation_action,
             ...
            }
    """
    transformers = {
        suffix: auto_kwargs(transformer) if transformer else transformer
        for suffix, transformer in transformers.items()
    }
    default = default and auto_kwargs(default)

    def transform(data, request):
        path = request.path
        transformer = default
        for suffix_test, suffix_transformer in transformers.items():
            if path.endswith(suffix_test):
                transformer = suffix_transformer
                break

        return transformer(data) if transformer else data

    return transform


def prefix(transformers, default=None):
    """Returns a different transformer depending on the prefix at the end of the requested URL.
       If none match and no default is given no transformation takes place.

       should pass in a dict with the following format:

            {'[prefix]': transformation_action,
             ...
            }
    """
    transformers = {
        prefix: auto_kwargs(transformer) if transformer else transformer
        for prefix, transformer in transformers.items()
    }
    default = default and auto_kwargs(default)

    def transform(data, request=None, response=None):
        path = request.path
        transformer = default
        for prefix_test, prefix_transformer in transformers.items():
            if path.startswith(prefix_test):
                transformer = prefix_transformer
                break

        return transformer(data) if transformer else data

    return transform


def all(*transformers):
    """Returns the results of applying all passed in transformers to data

       should pass in list of transformers

            [transformer_1, transformer_2...]
    """
    transformers = tuple(auto_kwargs(transformer) for transformer in transformers)

    def transform(data, request=None, response=None):
        for transformer in transformers:
            data = transformer(data, request=request, response=response)

        return data

    return transform



"""hug/route.py

Defines user usable routers

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import

from functools import partial
from types import FunctionType, MethodType

from falcon import HTTP_METHODS

import hug.api
from hug.routing import CLIRouter as cli  # noqa:  N813
from hug.routing import ExceptionRouter as exception  # noqa:  N813
from hug.routing import LocalRouter as local  # noqa:  N813
from hug.routing import NotFoundRouter as not_found  # noqa:  N813
from hug.routing import SinkRouter as sink  # noqa:  N813
from hug.routing import StaticRouter as static  # noqa:  N813
from hug.routing import URLRouter as http  # noqa:  N813


class Object(http):
    """Defines a router for classes and objects"""

    def __init__(self, urls=None, accept=HTTP_METHODS, output=None, **kwargs):
        super().__init__(urls=urls, accept=accept, output=output, **kwargs)

    def __call__(self, method_or_class=None, **kwargs):
        if not method_or_class and kwargs:
            return self.where(**kwargs)

        if isinstance(method_or_class, (MethodType, FunctionType)):
            routes = getattr(method_or_class, "_hug_http_routes", [])
            routes.append(self.route)
            method_or_class._hug_http_routes = routes
            return method_or_class

        instance = method_or_class
        if isinstance(method_or_class, type):
            instance = method_or_class()

        for argument in dir(instance):
            argument = getattr(instance, argument, None)

            http_routes = getattr(argument, "_hug_http_routes", ())
            for route in http_routes:
                http(**self.where(**route).route)(argument)

            cli_routes = getattr(argument, "_hug_cli_routes", ())
            for route in cli_routes:
                cli(**self.where(**route).route)(argument)

        return method_or_class

    def http_methods(self, urls=None, **route_data):
        """Creates routes from a class, where the class method names should line up to HTTP METHOD types"""

        def decorator(class_definition):
            instance = class_definition
            if isinstance(class_definition, type):
                instance = class_definition()

            router = self.urls(
                urls if urls else "/{0}".format(instance.__class__.__name__.lower()), **route_data
            )
            for method in HTTP_METHODS:
                handler = getattr(instance, method.lower(), None)
                if handler:
                    http_routes = getattr(handler, "_hug_http_routes", ())
                    if http_routes:
                        for route in http_routes:
                            http(**router.accept(method).where(**route).route)(handler)
                    else:
                        http(**router.accept(method).route)(handler)

                    cli_routes = getattr(handler, "_hug_cli_routes", ())
                    if cli_routes:
                        for route in cli_routes:
                            cli(**self.where(**route).route)(handler)
            return class_definition

        return decorator

    def cli(self, method):
        """Registers a method on an Object as a CLI route"""
        routes = getattr(method, "_hug_cli_routes", [])
        routes.append(self.route)
        method._hug_cli_routes = routes
        return method


class API(object):
    """Provides a convient way to route functions to a single API independent of where they live"""

    __slots__ = ("api",)

    def __init__(self, api):
        if type(api) == str:
            api = hug.api.API(api)
        self.api = api

    def http(self, *args, **kwargs):
        """Starts the process of building a new HTTP route linked to this API instance"""
        kwargs["api"] = self.api
        return http(*args, **kwargs)

    def urls(self, *args, **kwargs):
        """DEPRECATED: for backwords compatibility with < hug 2.2.0. `API.http` should be used instead.

           Starts the process of building a new URL HTTP route linked to this API instance
        """
        return self.http(*args, **kwargs)

    def not_found(self, *args, **kwargs):
        """Defines the handler that should handle not found requests against this API"""
        kwargs["api"] = self.api
        return not_found(*args, **kwargs)

    def static(self, *args, **kwargs):
        """Define the routes to static files the API should expose"""
        kwargs["api"] = self.api
        return static(*args, **kwargs)

    def sink(self, *args, **kwargs):
        """Define URL prefixes/handler matches where everything under the URL prefix should be handled"""
        kwargs["api"] = self.api
        return sink(*args, **kwargs)

    def exception(self, *args, **kwargs):
        """Defines how this API should handle the provided exceptions"""
        kwargs["api"] = self.api
        return exception(*args, **kwargs)

    def cli(self, *args, **kwargs):
        """Defines a CLI function that should be routed by this API"""
        kwargs["api"] = self.api
        return cli(*args, **kwargs)

    def object(self, *args, **kwargs):
        """Registers a class based router to this API"""
        kwargs["api"] = self.api
        return Object(*args, **kwargs)

    def get(self, *args, **kwargs):
        """Builds a new GET HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("GET",)
        return http(*args, **kwargs)

    def post(self, *args, **kwargs):
        """Builds a new POST HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("POST",)
        return http(*args, **kwargs)

    def put(self, *args, **kwargs):
        """Builds a new PUT HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("PUT",)
        return http(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """Builds a new DELETE HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("DELETE",)
        return http(*args, **kwargs)

    def connect(self, *args, **kwargs):
        """Builds a new CONNECT HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("CONNECT",)
        return http(*args, **kwargs)

    def head(self, *args, **kwargs):
        """Builds a new HEAD HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("HEAD",)
        return http(*args, **kwargs)

    def options(self, *args, **kwargs):
        """Builds a new OPTIONS HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("OPTIONS",)
        return http(*args, **kwargs)

    def patch(self, *args, **kwargs):
        """Builds a new PATCH HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("PATCH",)
        return http(*args, **kwargs)

    def trace(self, *args, **kwargs):
        """Builds a new TRACE HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("TRACE",)
        return http(*args, **kwargs)

    def get_post(self, *args, **kwargs):
        """Builds a new GET or POST HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("GET", "POST")
        return http(*args, **kwargs)

    def put_post(self, *args, **kwargs):
        """Builds a new PUT or POST HTTP route that is registered to this API"""
        kwargs["api"] = self.api
        kwargs["accept"] = ("PUT", "POST")
        return http(*args, **kwargs)


for method in HTTP_METHODS:
    method_handler = partial(http, accept=(method,))
    method_handler.__doc__ = "Exposes a Python method externally as an HTTP {0} method".format(
        method.upper()
    )
    globals()[method.lower()] = method_handler

get_post = partial(http, accept=("GET", "POST"))
get_post.__doc__ = "Exposes a Python method externally under both the HTTP POST and GET methods"

put_post = partial(http, accept=("PUT", "POST"))
put_post.__doc__ = "Exposes a Python method externally under both the HTTP POST and PUT methods"

object = Object()

# DEPRECATED: for backwords compatibility with hug 1.x.x
call = http



#!/usr/bin/env python
"""setup.py

Defines the setup instructions for the hug framework

Copyright (C) 2016  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
import glob
import os
import sys
from os import path

from setuptools import Extension, setup

MYDIR = path.abspath(os.path.dirname(__file__))
CYTHON = False
JYTHON = "java" in sys.platform

ext_modules = []
cmdclass = {}

try:
    sys.pypy_version_info
    PYPY = True
except AttributeError:
    PYPY = False

if not PYPY and not JYTHON:
    if "--without-cython" in sys.argv:
        sys.argv.remove("--without-cython")
        CYTHON = False
    else:
        try:
            from Cython.Distutils import build_ext

            CYTHON = True
        except ImportError:
            CYTHON = False

if CYTHON:

    def list_modules(dirname):
        filenames = glob.glob(path.join(dirname, "*.py"))

        module_names = []
        for name in filenames:
            module, ext = path.splitext(path.basename(name))
            if module != "__init__":
                module_names.append(module)

        return module_names

    ext_modules = [
        Extension("hug." + ext, [path.join("hug", ext + ".py")])
        for ext in list_modules(path.join(MYDIR, "hug"))
    ]
    cmdclass["build_ext"] = build_ext


with open("README.md", encoding="utf-8") as f:  # Loads in the README for PyPI
    long_description = f.read()


setup(
    name="hug",
    version="2.5.5",
    description="A Python framework that makes developing APIs "
    "as simple as possible, but no simpler.",
    long_description=long_description,
    # PEP 566, the new PyPI, and setuptools>=38.6.0 make markdown possible
    long_description_content_type="text/markdown",
    author="Timothy Crosley",
    author_email="timothy.crosley@gmail.com",
    # These appear in the left hand side bar on PyPI
    url="https://github.com/timothycrosley/hug",
    project_urls={
        "Documentation": "http://www.hug.rest/",
        "Gitter": "https://gitter.im/timothycrosley/hug",
    },
    license="MIT",
    entry_points={"console_scripts": ["hug = hug:development_runner.hug.interface.cli"]},
    packages=["hug"],
    requires=["falcon", "requests"],
    install_requires=["falcon==2.0.0", "requests"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "mock", "marshmallow"],
    ext_modules=ext_modules,
    cmdclass=cmdclass,
    python_requires=">=3.5",
    keywords="Web, Python, Python3, Refactoring, REST, Framework, RPC",
    classifiers=[
        "Development Status :: 6 - Mature",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)
